@@ -6,6 +6,7 @@
 
 #include "mozilla/dom/cache/CacheStorage.h"
 
+#include "mozilla/Preferences.h"
 #include "mozilla/Unused.h"
 #include "mozilla/dom/CacheBinding.h"
 #include "mozilla/dom/CacheStorageBinding.h"
@@ -15,6 +16,7 @@
 #include "mozilla/dom/cache/AutoUtils.h"
 #include "mozilla/dom/cache/Cache.h"
 #include "mozilla/dom/cache/CacheChild.h"
+#include "mozilla/dom/cache/CacheCommon.h"
 #include "mozilla/dom/cache/CacheStorageChild.h"
 #include "mozilla/dom/cache/CacheWorkerRef.h"
 #include "mozilla/dom/cache/PCacheChild.h"
@@ -33,10 +35,10 @@
 #include "nsIGlobalObject.h"
 #include "nsMixedContentBlocker.h"
 #include "nsURLParsers.h"
+#include "js/Object.h"              // JS::GetClass
+#include "js/PropertyAndElement.h"  // JS_DefineProperty
 
-namespace mozilla {
-namespace dom {
-namespace cache {
+namespace mozilla::dom::cache {
 
 using mozilla::ErrorResult;
 using mozilla::Unused;
@@ -77,12 +79,10 @@ bool IsTrusted(const PrincipalInfo& aPrincipalInfo, bool aTestingPrefEnabled) {
   }
 
   // Require a ContentPrincipal to avoid null principal, etc.
-  if (NS_WARN_IF(aPrincipalInfo.type() !=
-                 PrincipalInfo::TContentPrincipalInfo)) {
-    return false;
-  }
+  QM_TRY(OkIf(aPrincipalInfo.type() == PrincipalInfo::TContentPrincipalInfo),
+         false);
 
-  // If we're in testing mode, then don't do any more work to determing if
+  // If we're in testing mode, then don't do any more work to determine if
   // the origin is trusted.  We have to run some tests as http.
   if (aTestingPrefEnabled) {
     return true;
@@ -97,48 +97,38 @@ bool IsTrusted(const PrincipalInfo& aPrincipalInfo, bool aTestingPrefEnabled) {
   // TODO: Implement full secure setting algorithm. (bug 1177856)
 
   const nsCString& flatURL = aPrincipalInfo.get_ContentPrincipalInfo().spec();
-  const char* url = flatURL.get();
+  const char* const url = flatURL.get();
 
   // off the main thread URL parsing using nsStdURLParser.
-  nsCOMPtr<nsIURLParser> urlParser = new nsStdURLParser();
+  const nsCOMPtr<nsIURLParser> urlParser = new nsStdURLParser();
 
   uint32_t schemePos;
   int32_t schemeLen;
   uint32_t authPos;
   int32_t authLen;
-  nsresult rv =
+  QM_TRY(
       urlParser->ParseURL(url, flatURL.Length(), &schemePos, &schemeLen,
-                          &authPos, &authLen, nullptr, nullptr);  // ignore path
-  if (NS_WARN_IF(NS_FAILED(rv))) {
-    return false;
-  }
+                          &authPos, &authLen, nullptr, nullptr),  // ignore path
+      false);
 
-  nsAutoCString scheme(Substring(flatURL, schemePos, schemeLen));
+  const nsAutoCString scheme(Substring(flatURL, schemePos, schemeLen));
   if (scheme.LowerCaseEqualsLiteral("https") ||
-      scheme.LowerCaseEqualsLiteral("file")) {
-    return true;
-  }
-
-  if (StaticPrefs::extensions_backgroundServiceWorker_enabled_AtStartup() &&
+      scheme.LowerCaseEqualsLiteral("file") ||
       scheme.LowerCaseEqualsLiteral("moz-extension")) {
     return true;
   }
 
   uint32_t hostPos;
   int32_t hostLen;
+  QM_TRY(urlParser->ParseAuthority(url + authPos, authLen, nullptr,
+                                   nullptr,           // ignore username
+                                   nullptr, nullptr,  // ignore password
+                                   &hostPos, &hostLen,
+                                   nullptr),  // ignore port
+         false);
 
-  rv = urlParser->ParseAuthority(url + authPos, authLen, nullptr,
-                                 nullptr,           // ignore username
-                                 nullptr, nullptr,  // ignore password
-                                 &hostPos, &hostLen,
-                                 nullptr);  // ignore port
-  if (NS_WARN_IF(NS_FAILED(rv))) {
-    return false;
-  }
-
-  nsDependentCSubstring hostname(url + authPos + hostPos, hostLen);
-
-  return nsMixedContentBlocker::IsPotentiallyTrustworthyLoopbackHost(hostname);
+  return nsMixedContentBlocker::IsPotentiallyTrustworthyLoopbackHost(
+      nsDependentCSubstring(url + authPos + hostPos, hostLen));
 }
 
 }  // namespace
@@ -152,19 +142,16 @@ already_AddRefed<CacheStorage> CacheStorage::CreateOnMainThread(
   MOZ_ASSERT(NS_IsMainThread());
 
   PrincipalInfo principalInfo;
-  nsresult rv = PrincipalToPrincipalInfo(aPrincipal, &principalInfo);
-  if (NS_WARN_IF(NS_FAILED(rv))) {
-    aRv.Throw(rv);
-    return nullptr;
-  }
+  QM_TRY(PrincipalToPrincipalInfo(aPrincipal, &principalInfo), nullptr,
+         [&aRv](const nsresult rv) { aRv.Throw(rv); });
 
-  if (NS_WARN_IF(!QuotaManager::IsPrincipalInfoValid(principalInfo))) {
-    NS_WARNING("CacheStorage not supported on invalid origins.");
-    RefPtr<CacheStorage> ref = new CacheStorage(NS_ERROR_DOM_SECURITY_ERR);
-    return ref.forget();
-  }
+  QM_TRY(OkIf(QuotaManager::IsPrincipalInfoValid(principalInfo)),
+         RefPtr{new CacheStorage(NS_ERROR_DOM_SECURITY_ERR)}.forget(),
+         [](const auto) {
+           NS_WARNING("CacheStorage not supported on invalid origins.");
+         });
 
-  bool testingEnabled =
+  const bool testingEnabled =
       aForceTrustedOrigin ||
       Preferences::GetBool("dom.caches.testing.enabled", false) ||
       StaticPrefs::dom_serviceWorkers_testing_enabled();
@@ -205,10 +192,8 @@ already_AddRefed<CacheStorage> CacheStorage::CreateOnWorker(
   const PrincipalInfo& principalInfo =
       aWorkerPrivate->GetEffectiveStoragePrincipalInfo();
 
-  if (NS_WARN_IF(!QuotaManager::IsPrincipalInfoValid(principalInfo))) {
-    aRv.Throw(NS_ERROR_FAILURE);
-    return nullptr;
-  }
+  QM_TRY(OkIf(QuotaManager::IsPrincipalInfoValid(principalInfo)), nullptr,
+         [&aRv](const auto) { aRv.Throw(NS_ERROR_FAILURE); });
 
   // We have a number of cases where we want to skip the https scheme
   // validation:
@@ -242,7 +227,7 @@ already_AddRefed<CacheStorage> CacheStorage::CreateOnWorker(
 // static
 bool CacheStorage::DefineCaches(JSContext* aCx, JS::Handle<JSObject*> aGlobal) {
   MOZ_ASSERT(NS_IsMainThread());
-  MOZ_DIAGNOSTIC_ASSERT(js::GetObjectClass(aGlobal)->flags & JSCLASS_DOM_GLOBAL,
+  MOZ_DIAGNOSTIC_ASSERT(JS::GetClass(aGlobal)->flags & JSCLASS_DOM_GLOBAL,
                         "Passed object is not a global object!");
   js::AssertSameCompartment(aCx, aGlobal);
 
@@ -313,7 +298,7 @@ CacheStorage::CacheStorage(nsresult aFailureResult)
 
 already_AddRefed<Promise> CacheStorage::Match(
     JSContext* aCx, const RequestOrUSVString& aRequest,
-    const CacheQueryOptions& aOptions, ErrorResult& aRv) {
+    const MultiCacheQueryOptions& aOptions, ErrorResult& aRv) {
   NS_ASSERT_OWNINGTHREAD(CacheStorage);
 
   if (!HasStorageAccess()) {
@@ -590,6 +575,4 @@ bool CacheStorage::HasStorageAccess() const {
   return access > StorageAccess::ePrivateBrowsing;
 }
 
-}  // namespace cache
-}  // namespace dom
-}  // namespace mozilla
+}  // namespace mozilla::dom::cache
