@@ -14,12 +14,13 @@
 #include "nsWrapperCache.h"
 #include "mozilla/dom/WebGLRenderingContextBinding.h"
 #include "mozilla/dom/WebGL2RenderingContextBinding.h"
+#include "mozilla/layers/LayersSurfaces.h"
+#include "mozilla/StaticPrefs_webgl.h"
 #include "WebGLFormats.h"
 #include "WebGLStrongTypes.h"
 #include "WebGLTypes.h"
 
 #include "mozilla/Logging.h"
-#include "WebGLCrossProcessCommandQueue.h"
 #include "WebGLCommandQueue.h"
 
 #include <memory>
@@ -27,19 +28,10 @@
 #include <unordered_set>
 #include <vector>
 
-#ifndef WEBGL_BRIDGE_LOG_
-#  define WEBGL_BRIDGE_LOG_(lvl, ...) \
-    MOZ_LOG(mozilla::gWebGLBridgeLog, lvl, (__VA_ARGS__))
-#  define WEBGL_BRIDGE_LOGV(...) \
-    WEBGL_BRIDGE_LOG_(LogLevel::Verbose, __VA_ARGS__)
-#  define WEBGL_BRIDGE_LOGD(...) WEBGL_BRIDGE_LOG_(LogLevel::Debug, __VA_ARGS__)
-#  define WEBGL_BRIDGE_LOGI(...) WEBGL_BRIDGE_LOG_(LogLevel::Info, __VA_ARGS__)
-#  define WEBGL_BRIDGE_LOGE(...) WEBGL_BRIDGE_LOG_(LogLevel::Error, __VA_ARGS__)
-#endif  // WEBGL_BRIDGE_LOG_
-
 namespace mozilla {
 
 class ClientWebGLExtensionBase;
+class HostWebGLContext;
 
 namespace dom {
 class WebGLChild;
@@ -69,9 +61,7 @@ class WebGLActiveInfoJS final : public RefCounted<WebGLActiveInfoJS> {
   GLint Size() const { return static_cast<GLint>(mInfo.elemCount); }
   GLenum Type() const { return mInfo.elemType; }
 
-  void GetName(nsString& retval) const {
-    retval = NS_ConvertUTF8toUTF16(mInfo.name.c_str());
-  }
+  void GetName(nsString& retval) const { CopyUTF8toUTF16(mInfo.name, retval); }
 
   bool WrapObject(JSContext*, JS::Handle<JSObject*>,
                   JS::MutableHandle<JSObject*>);
@@ -169,7 +159,6 @@ class ContextGenerationInfo final {
 
   std::vector<TypedQuad> mGenericVertexAttribs;
 
-  std::array<bool, 4> mColorWriteMask = {{true, true, true, true}};
   std::array<int32_t, 4> mScissor = {};
   std::array<int32_t, 4> mViewport = {};
   std::array<float, 4> mClearColor = {{0, 0, 0, 0}};
@@ -187,24 +176,19 @@ class ContextGenerationInfo final {
 
 // -
 
-struct RemotingData final {
-  // In the cross process case, the WebGL actor's ownership relationship looks
-  // like this:
-  // ---------------------------------------------------------------------
-  // | ClientWebGLContext -> WebGLChild -> WebGLParent -> HostWebGLContext
-  // ---------------------------------------------------------------------
-  //
-  // where 'A -> B' means "A owns B"
-  RefPtr<mozilla::dom::WebGLChild> mWebGLChild;
-  UniquePtr<ClientWebGLCommandSourceP> mCommandSourcePcq;
-  UniquePtr<ClientWebGLCommandSourceI> mCommandSourceIpdl;
-};
+// In the cross process case, the WebGL actor's ownership relationship looks
+// like this:
+// ---------------------------------------------------------------------
+// | ClientWebGLContext -> WebGLChild -> WebGLParent -> HostWebGLContext
+// ---------------------------------------------------------------------
+//
+// where 'A -> B' means "A owns B"
 
 struct NotLostData final {
   ClientWebGLContext& context;
   webgl::InitContextResult info;
 
-  Maybe<RemotingData> outOfProcess;
+  RefPtr<mozilla::dom::WebGLChild> outOfProcess;
   UniquePtr<HostWebGLContext> inProcess;
 
   webgl::ContextGenerationInfo state;
@@ -614,9 +598,9 @@ class WebGLVertexArrayJS final : public nsWrapperCache, public webgl::ObjectJS {
 
 ////////////////////////////////////
 
-typedef dom::MaybeSharedFloat32ArrayOrUnrestrictedFloatSequence Float32ListU;
-typedef dom::MaybeSharedInt32ArrayOrLongSequence Int32ListU;
-typedef dom::MaybeSharedUint32ArrayOrUnsignedLongSequence Uint32ListU;
+using Float32ListU = dom::MaybeSharedFloat32ArrayOrUnrestrictedFloatSequence;
+using Int32ListU = dom::MaybeSharedInt32ArrayOrLongSequence;
+using Uint32ListU = dom::MaybeSharedUint32ArrayOrUnsignedLongSequence;
 
 inline Range<const float> MakeRange(const Float32ListU& list) {
   if (list.IsFloat32Array()) return MakeRangeAbv(list.GetAsFloat32Array());
@@ -794,7 +778,7 @@ class ClientWebGLContext final : public nsICanvasRenderingContextInternal,
   mozilla::dom::WebGLChild* GetChild() const {
     if (!mNotLost) return nullptr;
     if (!mNotLost->outOfProcess) return nullptr;
-    return mNotLost->outOfProcess->mWebGLChild.get();
+    return mNotLost->outOfProcess.get();
   }
 
   // -------------------------------------------------------------------------
@@ -878,6 +862,11 @@ class ClientWebGLContext final : public nsICanvasRenderingContextInternal,
     EnqueueError(0, format, args...);
   }
 
+  template <typename... Args>
+  void EnqueuePerfWarning(const char* const format, const Args&... args) const {
+    EnqueueError(webgl::kErrorPerfWarning, format, args...);
+  }
+
   void EnqueueError_ArgEnum(const char* argName,
                             GLenum val) const;  // Cold code.
 
@@ -1023,12 +1012,43 @@ class ClientWebGLContext final : public nsICanvasRenderingContextInternal,
   void DoReadPixels(const webgl::ReadPixelsDesc&, Range<uint8_t>) const;
   uvec2 DrawingBufferSize();
 
+  // -
+
+  bool mAutoFlushPending = false;
+
+  void AutoEnqueueFlush() {
+    if (MOZ_LIKELY(mAutoFlushPending)) return;
+    mAutoFlushPending = true;
+
+    const auto weak = WeakPtr<ClientWebGLContext>(this);
+    const auto DeferredFlush = [weak]() {
+      const auto strong = RefPtr<ClientWebGLContext>(weak);
+      if (!strong) return;
+      if (!strong->mAutoFlushPending) return;
+      strong->mAutoFlushPending = false;
+
+      if (!StaticPrefs::webgl_auto_flush()) return;
+      const bool flushGl = StaticPrefs::webgl_auto_flush_gl();
+      strong->Flush(flushGl);
+    };
+
+    already_AddRefed<mozilla::Runnable> runnable = NS_NewRunnableFunction(
+        "enqueue Event_webglcontextrestored", DeferredFlush);
+    NS_DispatchToCurrentThread(std::move(runnable));
+  }
+
+  void CancelAutoFlush() { mAutoFlushPending = false; }
+
+  // -
+
   void AfterDrawCall() {
     if (!mNotLost) return;
     const auto& state = State();
     if (!state.mBoundDrawFb) {
       MarkCanvasDirty();
     }
+
+    AutoEnqueueFlush();
   }
 
   // -------------------------------------------------------------------------
@@ -1039,8 +1059,9 @@ class ClientWebGLContext final : public nsICanvasRenderingContextInternal,
  public:
   bool IsContextLost() const { return !mNotLost; }
 
-  void Disable(GLenum cap) const;
-  void Enable(GLenum cap) const;
+  void Disable(GLenum cap) const { SetEnabledI(cap, {}, false); }
+  void Enable(GLenum cap) const { SetEnabledI(cap, {}, true); }
+  void SetEnabledI(GLenum cap, Maybe<GLuint> i, bool val) const;
   bool IsEnabled(GLenum cap) const;
 
  private:
@@ -1207,9 +1228,18 @@ class ClientWebGLContext final : public nsICanvasRenderingContextInternal,
     BlendFuncSeparate(sfactor, dfactor, sfactor, dfactor);
   }
 
-  void BlendEquationSeparate(GLenum modeRGB, GLenum modeAlpha);
+  void BlendEquationSeparate(GLenum modeRGB, GLenum modeAlpha) {
+    BlendEquationSeparateI({}, modeRGB, modeAlpha);
+  }
   void BlendFuncSeparate(GLenum srcRGB, GLenum dstRGB, GLenum srcAlpha,
-                         GLenum dstAlpha);
+                         GLenum dstAlpha) {
+    BlendFuncSeparateI({}, srcRGB, dstRGB, srcAlpha, dstAlpha);
+  }
+
+  void BlendEquationSeparateI(Maybe<GLuint> buf, GLenum modeRGB,
+                              GLenum modeAlpha);
+  void BlendFuncSeparateI(Maybe<GLuint> buf, GLenum srcRGB, GLenum dstRGB,
+                          GLenum srcAlpha, GLenum dstAlpha);
 
   // -
 
@@ -1251,8 +1281,10 @@ class ClientWebGLContext final : public nsICanvasRenderingContextInternal,
 
   void ClearStencil(GLint v);
 
-  void ColorMask(WebGLboolean r, WebGLboolean g, WebGLboolean b,
-                 WebGLboolean a);
+  void ColorMask(bool r, bool g, bool b, bool a) const {
+    ColorMaskI({}, r, g, b, a);
+  }
+  void ColorMaskI(Maybe<GLuint> buf, bool r, bool g, bool b, bool a) const;
 
   void CullFace(GLenum face);
 
@@ -1262,7 +1294,7 @@ class ClientWebGLContext final : public nsICanvasRenderingContextInternal,
 
   void DepthRange(GLclampf zNear, GLclampf zFar);
 
-  void Flush();
+  void Flush(bool flushGl = true);
 
   void Finish();
 
