@@ -6,20 +6,32 @@
 
 #include "mozilla/dom/ImageBitmap.h"
 #include "mozilla/CheckedInt.h"
+#include "mozilla/dom/BlobImpl.h"
+#include "mozilla/dom/CanvasRenderingContext2D.h"
+#include "mozilla/dom/CanvasUtils.h"
+#include "mozilla/dom/Document.h"
+#include "mozilla/dom/HTMLCanvasElement.h"
+#include "mozilla/dom/HTMLImageElement.h"
 #include "mozilla/dom/HTMLMediaElementBinding.h"
+#include "mozilla/dom/HTMLVideoElement.h"
 #include "mozilla/dom/ImageBitmapBinding.h"
+#include "mozilla/dom/OffscreenCanvas.h"
 #include "mozilla/dom/Promise.h"
 #include "mozilla/dom/StructuredCloneTags.h"
+#include "mozilla/dom/SVGImageElement.h"
 #include "mozilla/dom/WorkerPrivate.h"
 #include "mozilla/dom/WorkerRef.h"
 #include "mozilla/dom/WorkerRunnable.h"
 #include "mozilla/gfx/2D.h"
+#include "mozilla/gfx/Logging.h"
 #include "mozilla/gfx/Swizzle.h"
 #include "mozilla/Mutex.h"
 #include "mozilla/ScopeExit.h"
+#include "nsGlobalWindowInner.h"
+#include "nsIAsyncInputStream.h"
 #include "nsNetUtil.h"
+#include "nsLayoutUtils.h"
 #include "nsStreamUtils.h"
-#include "ImageUtils.h"
 #include "imgLoader.h"
 #include "imgTools.h"
 
@@ -28,8 +40,7 @@ using namespace mozilla::layers;
 using mozilla::dom::HTMLMediaElement_Binding::HAVE_METADATA;
 using mozilla::dom::HTMLMediaElement_Binding::NETWORK_EMPTY;
 
-namespace mozilla {
-namespace dom {
+namespace mozilla::dom {
 
 NS_IMPL_CYCLE_COLLECTION_WRAPPERCACHE(ImageBitmap, mParent)
 NS_IMPL_CYCLE_COLLECTING_ADDREF(ImageBitmap)
@@ -257,6 +268,59 @@ static already_AddRefed<DataSourceSurface> CropAndCopyDataSourceSurface(
   return dstDataSurface.forget();
 }
 
+static DataSourceSurface* FlipYDataSourceSurface(DataSourceSurface* aSurface) {
+  MOZ_ASSERT(aSurface);
+
+  // Invert in y direction.
+  DataSourceSurface::ScopedMap srcMap(aSurface, DataSourceSurface::READ_WRITE);
+  if (NS_WARN_IF(!srcMap.IsMapped())) {
+    return nullptr;
+  }
+
+  const IntSize srcSize = aSurface->GetSize();
+  uint8_t* srcBufferPtr = srcMap.GetData();
+  const uint32_t stride = srcMap.GetStride();
+
+  CheckedInt<uint32_t> copiedBytesPerRaw = CheckedInt<uint32_t>(stride);
+  if (!copiedBytesPerRaw.isValid()) {
+    return nullptr;
+  }
+
+  for (int i = 0; i < srcSize.height / 2; ++i) {
+    std::swap_ranges(srcBufferPtr + stride * i, srcBufferPtr + stride * (i + 1),
+                     srcBufferPtr + stride * (srcSize.height - 1 - i));
+  }
+
+  return aSurface;
+}
+
+static DataSourceSurface* AlphaPremultiplyDataSourceSurface(
+    DataSourceSurface* aSurface, const bool forward = true) {
+  MOZ_ASSERT(aSurface);
+
+  DataSourceSurface::MappedSurface surfaceMap;
+
+  if (aSurface->Map(DataSourceSurface::MapType::READ_WRITE, &surfaceMap)) {
+    if (forward) {
+      PremultiplyData(surfaceMap.mData, surfaceMap.mStride,
+                      aSurface->GetFormat(), surfaceMap.mData,
+                      surfaceMap.mStride, aSurface->GetFormat(),
+                      aSurface->GetSize());
+    } else {
+      UnpremultiplyData(surfaceMap.mData, surfaceMap.mStride,
+                        aSurface->GetFormat(), surfaceMap.mData,
+                        surfaceMap.mStride, aSurface->GetFormat(),
+                        aSurface->GetSize());
+    }
+
+    aSurface->Unmap();
+  } else {
+    return nullptr;
+  }
+
+  return aSurface;
+}
+
 /*
  * Encapsulate the given _aSurface_ into a layers::SourceSurfaceImage.
  */
@@ -275,7 +339,8 @@ static already_AddRefed<layers::Image> CreateImageFromSurface(
  */
 static already_AddRefed<SourceSurface> CreateSurfaceFromRawData(
     const gfx::IntSize& aSize, uint32_t aStride, gfx::SurfaceFormat aFormat,
-    uint8_t* aBuffer, uint32_t aBufferLength, const Maybe<IntRect>& aCropRect) {
+    uint8_t* aBuffer, uint32_t aBufferLength, const Maybe<IntRect>& aCropRect,
+    const ImageBitmapOptions& aOptions) {
   MOZ_ASSERT(!aSize.IsEmpty());
   MOZ_ASSERT(aBuffer);
 
@@ -301,17 +366,34 @@ static already_AddRefed<SourceSurface> CreateSurfaceFromRawData(
     return nullptr;
   }
 
+  if (aOptions.mImageOrientation == ImageOrientation::FlipY) {
+    result = FlipYDataSourceSurface(result);
+
+    if (NS_WARN_IF(!result)) {
+      return nullptr;
+    }
+  }
+
+  if (aOptions.mPremultiplyAlpha == PremultiplyAlpha::Premultiply) {
+    result = AlphaPremultiplyDataSourceSurface(result);
+
+    if (NS_WARN_IF(!result)) {
+      return nullptr;
+    }
+  }
+
   return result.forget();
 }
 
 static already_AddRefed<layers::Image> CreateImageFromRawData(
     const gfx::IntSize& aSize, uint32_t aStride, gfx::SurfaceFormat aFormat,
-    uint8_t* aBuffer, uint32_t aBufferLength, const Maybe<IntRect>& aCropRect) {
+    uint8_t* aBuffer, uint32_t aBufferLength, const Maybe<IntRect>& aCropRect,
+    const ImageBitmapOptions& aOptions) {
   MOZ_ASSERT(NS_IsMainThread());
 
   // Copy and crop the source buffer into a SourceSurface.
   RefPtr<SourceSurface> rgbaSurface = CreateSurfaceFromRawData(
-      aSize, aStride, aFormat, aBuffer, aBufferLength, aCropRect);
+      aSize, aStride, aFormat, aBuffer, aBufferLength, aCropRect, aOptions);
 
   if (NS_WARN_IF(!rgbaSurface)) {
     return nullptr;
@@ -344,13 +426,7 @@ static already_AddRefed<layers::Image> CreateImageFromRawData(
               bgraDataSurface->GetSize());
 
   // Create an Image from the BGRA SourceSurface.
-  RefPtr<layers::Image> image = CreateImageFromSurface(bgraDataSurface);
-
-  if (NS_WARN_IF(!image)) {
-    return nullptr;
-  }
-
-  return image.forget();
+  return CreateImageFromSurface(bgraDataSurface);
 }
 
 /*
@@ -370,7 +446,8 @@ class CreateImageFromRawDataInMainThreadSyncTask final
   CreateImageFromRawDataInMainThreadSyncTask(
       uint8_t* aBuffer, uint32_t aBufferLength, uint32_t aStride,
       gfx::SurfaceFormat aFormat, const gfx::IntSize& aSize,
-      const Maybe<IntRect>& aCropRect, layers::Image** aImage)
+      const Maybe<IntRect>& aCropRect, layers::Image** aImage,
+      const ImageBitmapOptions& aOptions)
       : WorkerMainThreadRunnable(
             GetCurrentThreadWorkerPrivate(),
             "ImageBitmap :: Create Image from Raw Data"_ns),
@@ -380,7 +457,8 @@ class CreateImageFromRawDataInMainThreadSyncTask final
         mStride(aStride),
         mFormat(aFormat),
         mSize(aSize),
-        mCropRect(aCropRect) {
+        mCropRect(aCropRect),
+        mOptions(aOptions) {
     MOZ_ASSERT(!(*aImage),
                "Don't pass an existing Image into "
                "CreateImageFromRawDataInMainThreadSyncTask.");
@@ -388,7 +466,7 @@ class CreateImageFromRawDataInMainThreadSyncTask final
 
   bool MainThreadRun() override {
     RefPtr<layers::Image> image = CreateImageFromRawData(
-        mSize, mStride, mFormat, mBuffer, mBufferLength, mCropRect);
+        mSize, mStride, mFormat, mBuffer, mBufferLength, mCropRect, mOptions);
 
     if (NS_WARN_IF(!image)) {
       return false;
@@ -407,6 +485,7 @@ class CreateImageFromRawDataInMainThreadSyncTask final
   gfx::SurfaceFormat mFormat;
   gfx::IntSize mSize;
   const Maybe<IntRect>& mCropRect;
+  const ImageBitmapOptions mOptions;
 };
 
 /*
@@ -416,10 +495,18 @@ class CreateImageFromRawDataInMainThreadSyncTask final
 template <class ElementType>
 static already_AddRefed<SourceSurface> GetSurfaceFromElement(
     nsIGlobalObject* aGlobal, ElementType& aElement, bool* aWriteOnly,
+    const ImageBitmapOptions& aOptions, gfxAlphaType* aAlphaType,
     ErrorResult& aRv) {
-  nsLayoutUtils::SurfaceFromElementResult res =
-      nsLayoutUtils::SurfaceFromElement(
-          &aElement, nsLayoutUtils::SFE_WANT_FIRST_FRAME_IF_IMAGE);
+  uint32_t flags = nsLayoutUtils::SFE_WANT_FIRST_FRAME_IF_IMAGE;
+
+  // by default surfaces have premultiplied alpha
+  // attempt to get non premultiplied if required
+  if (aOptions.mPremultiplyAlpha == PremultiplyAlpha::None) {
+    flags |= nsLayoutUtils::SFE_ALLOW_NON_PREMULT;
+  }
+
+  SurfaceFromElementResult res =
+      nsLayoutUtils::SurfaceFromElement(&aElement, flags);
 
   RefPtr<SourceSurface> surface = res.GetSourceSurface();
   if (NS_WARN_IF(!surface)) {
@@ -428,6 +515,7 @@ static already_AddRefed<SourceSurface> GetSurfaceFromElement(
   }
 
   *aWriteOnly = res.mIsWriteOnly;
+  *aAlphaType = res.mAlphaType;
 
   return surface.forget();
 }
@@ -437,7 +525,6 @@ ImageBitmap::ImageBitmap(nsIGlobalObject* aGlobal, layers::Image* aData,
     : mParent(aGlobal),
       mData(aData),
       mSurface(nullptr),
-      mDataWrapper(new ImageUtils(mData)),
       mPictureRect(aData->GetPictureRect()),
       mAlphaType(aAlphaType),
       mAllocatedImageData(false),
@@ -464,7 +551,6 @@ JSObject* ImageBitmap::WrapObject(JSContext* aCx,
 void ImageBitmap::Close() {
   mData = nullptr;
   mSurface = nullptr;
-  mDataWrapper = nullptr;
   mPictureRect.SetEmpty();
 }
 
@@ -587,6 +673,7 @@ already_AddRefed<SourceSurface> ImageBitmap::PrepareForDrawTarget(
       srcSurface->Unmap();
     }
 
+    mAlphaType = gfxAlphaType::Premult;
     mSurface = dstSurface;
   }
 
@@ -660,9 +747,8 @@ already_AddRefed<ImageBitmap> ImageBitmap::CreateFromOffscreenCanvas(
   // Check write-only mode.
   bool writeOnly = aOffscreenCanvas.IsWriteOnly();
 
-  nsLayoutUtils::SurfaceFromElementResult res =
-      nsLayoutUtils::SurfaceFromOffscreenCanvas(
-          &aOffscreenCanvas, nsLayoutUtils::SFE_WANT_FIRST_FRAME_IF_IMAGE);
+  SurfaceFromElementResult res = nsLayoutUtils::SurfaceFromOffscreenCanvas(
+      &aOffscreenCanvas, nsLayoutUtils::SFE_WANT_FIRST_FRAME_IF_IMAGE);
 
   RefPtr<SourceSurface> surface = res.GetSourceSurface();
 
@@ -680,10 +766,111 @@ already_AddRefed<ImageBitmap> ImageBitmap::CreateFromOffscreenCanvas(
   return ret.forget();
 }
 
+/* static */
+already_AddRefed<ImageBitmap> ImageBitmap::CreateImageBitmapInternal(
+    nsIGlobalObject* aGlobal, gfx::SourceSurface* aSurface,
+    const Maybe<IntRect>& aCropRect, const ImageBitmapOptions& aOptions,
+    const bool aWriteOnly, const bool aAllocatedImageData, const bool aMustCopy,
+    const gfxAlphaType aAlphaType, ErrorResult& aRv) {
+  bool needToReportMemoryAllocation = aAllocatedImageData;
+  const IntSize srcSize = aSurface->GetSize();
+  IntRect cropRect =
+      aCropRect.valueOr(IntRect(0, 0, srcSize.width, srcSize.height));
+
+  RefPtr<SourceSurface> surface = aSurface;
+  RefPtr<DataSourceSurface> dataSurface;
+
+  // handle alpha premultiplication if surface not of correct type
+
+  gfxAlphaType alphaType = aAlphaType;
+  bool mustCopy = aMustCopy;
+  bool requiresPremultiply = false;
+  bool requiresUnpremultiply = false;
+
+  if (!IsOpaque(surface->GetFormat())) {
+    if (aAlphaType == gfxAlphaType::Premult &&
+        aOptions.mPremultiplyAlpha == PremultiplyAlpha::None) {
+      requiresUnpremultiply = true;
+      alphaType = gfxAlphaType::NonPremult;
+      if (!aAllocatedImageData) {
+        mustCopy = true;
+      }
+    } else if (aAlphaType == gfxAlphaType::NonPremult &&
+               aOptions.mPremultiplyAlpha == PremultiplyAlpha::Premultiply) {
+      requiresPremultiply = true;
+      alphaType = gfxAlphaType::Premult;
+      if (!aAllocatedImageData) {
+        mustCopy = true;
+      }
+    }
+  }
+
+  /*
+   * if we don't own the data and need to create a new buffer to flip Y.
+   * or
+   * we need to crop and flip, where crop must come first.
+   * or
+   * Or the caller demands a copy (WebGL contexts).
+   */
+  if ((aOptions.mImageOrientation == ImageOrientation::FlipY &&
+       (!aAllocatedImageData || aCropRect.isSome())) ||
+      mustCopy) {
+    dataSurface = surface->GetDataSurface();
+
+    dataSurface = CropAndCopyDataSourceSurface(dataSurface, cropRect);
+    if (NS_WARN_IF(!dataSurface)) {
+      return nullptr;
+    }
+
+    surface = dataSurface;
+    cropRect.MoveTo(0, 0);
+    needToReportMemoryAllocation = true;
+  }
+
+  // flip image in Y direction
+  if (aOptions.mImageOrientation == ImageOrientation::FlipY) {
+    if (!dataSurface) {
+      dataSurface = surface->GetDataSurface();
+    }
+
+    surface = FlipYDataSourceSurface(dataSurface);
+    if (NS_WARN_IF(!surface)) {
+      return nullptr;
+    }
+  }
+
+  if (requiresPremultiply || requiresUnpremultiply) {
+    if (!dataSurface) {
+      dataSurface = surface->GetDataSurface();
+    }
+
+    surface =
+        AlphaPremultiplyDataSourceSurface(dataSurface, requiresPremultiply);
+    if (NS_WARN_IF(!surface)) {
+      return nullptr;
+    }
+  }
+
+  // Create an Image from the SourceSurface.
+  RefPtr<layers::Image> data = CreateImageFromSurface(surface);
+  RefPtr<ImageBitmap> ret =
+      new ImageBitmap(aGlobal, data, aWriteOnly, alphaType);
+
+  if (needToReportMemoryAllocation) {
+    ret->mAllocatedImageData = true;
+  }
+
+  // Set the picture rectangle.
+  ret->SetPictureRect(cropRect, aRv);
+
+  return ret.forget();
+}
+
 /* static */
 already_AddRefed<ImageBitmap> ImageBitmap::CreateInternal(
     nsIGlobalObject* aGlobal, HTMLImageElement& aImageEl,
-    const Maybe<IntRect>& aCropRect, ErrorResult& aRv) {
+    const Maybe<IntRect>& aCropRect, const ImageBitmapOptions& aOptions,
+    ErrorResult& aRv) {
   // Check if the image element is completely available or not.
   if (!aImageEl.Complete()) {
     aRv.Throw(NS_ERROR_DOM_INVALID_STATE_ERR);
@@ -691,71 +878,51 @@ already_AddRefed<ImageBitmap> ImageBitmap::CreateInternal(
   }
 
   bool writeOnly = true;
+  gfxAlphaType alphaType = gfxAlphaType::NonPremult;
 
   // Get the SourceSurface out from the image element and then do security
   // checking.
-  RefPtr<SourceSurface> surface =
-      GetSurfaceFromElement(aGlobal, aImageEl, &writeOnly, aRv);
+  RefPtr<SourceSurface> surface = GetSurfaceFromElement(
+      aGlobal, aImageEl, &writeOnly, aOptions, &alphaType, aRv);
 
   if (NS_WARN_IF(aRv.Failed())) {
     return nullptr;
   }
 
-  // Create ImageBitmap.
-  RefPtr<layers::Image> data = CreateImageFromSurface(surface);
-
-  if (NS_WARN_IF(!data)) {
-    aRv.Throw(NS_ERROR_NOT_AVAILABLE);
-    return nullptr;
-  }
-
-  RefPtr<ImageBitmap> ret = new ImageBitmap(aGlobal, data, writeOnly);
-
-  // Set the picture rectangle.
-  if (ret && aCropRect.isSome()) {
-    ret->SetPictureRect(aCropRect.ref(), aRv);
-  }
-
-  return ret.forget();
+  bool needToReportMemoryAllocation = false;
+  return CreateImageBitmapInternal(aGlobal, surface, aCropRect, aOptions,
+                                   writeOnly, needToReportMemoryAllocation,
+                                   false, alphaType, aRv);
 }
-
 /* static */
 already_AddRefed<ImageBitmap> ImageBitmap::CreateInternal(
     nsIGlobalObject* aGlobal, SVGImageElement& aImageEl,
-    const Maybe<IntRect>& aCropRect, ErrorResult& aRv) {
+    const Maybe<IntRect>& aCropRect, const ImageBitmapOptions& aOptions,
+    ErrorResult& aRv) {
   bool writeOnly = true;
+  gfxAlphaType alphaType = gfxAlphaType::NonPremult;
 
   // Get the SourceSurface out from the image element and then do security
   // checking.
-  RefPtr<SourceSurface> surface =
-      GetSurfaceFromElement(aGlobal, aImageEl, &writeOnly, aRv);
+  RefPtr<SourceSurface> surface = GetSurfaceFromElement(
+      aGlobal, aImageEl, &writeOnly, aOptions, &alphaType, aRv);
 
   if (NS_WARN_IF(aRv.Failed())) {
     return nullptr;
   }
 
-  // Create ImageBitmap.
-  RefPtr<layers::Image> data = CreateImageFromSurface(surface);
+  bool needToReportMemoryAllocation = false;
 
-  if (NS_WARN_IF(!data)) {
-    aRv.Throw(NS_ERROR_NOT_AVAILABLE);
-    return nullptr;
-  }
-
-  RefPtr<ImageBitmap> ret = new ImageBitmap(aGlobal, data, writeOnly);
-
-  // Set the picture rectangle.
-  if (ret && aCropRect.isSome()) {
-    ret->SetPictureRect(aCropRect.ref(), aRv);
-  }
-
-  return ret.forget();
+  return CreateImageBitmapInternal(aGlobal, surface, aCropRect, aOptions,
+                                   writeOnly, needToReportMemoryAllocation,
+                                   false, alphaType, aRv);
 }
 
 /* static */
 already_AddRefed<ImageBitmap> ImageBitmap::CreateInternal(
     nsIGlobalObject* aGlobal, HTMLVideoElement& aVideoEl,
-    const Maybe<IntRect>& aCropRect, ErrorResult& aRv) {
+    const Maybe<IntRect>& aCropRect, const ImageBitmapOptions& aOptions,
+    ErrorResult& aRv) {
   aVideoEl.MarkAsContentSource(
       mozilla::dom::HTMLVideoElement::CallerAPI::CREATE_IMAGEBITMAP);
 
@@ -777,8 +944,8 @@ already_AddRefed<ImageBitmap> ImageBitmap::CreateInternal(
   nsCOMPtr<nsIPrincipal> principal = aVideoEl.GetCurrentVideoPrincipal();
   bool hadCrossOriginRedirects = aVideoEl.HadCrossOriginRedirects();
   bool CORSUsed = aVideoEl.GetCORSMode() != CORS_NONE;
-  bool writeOnly =
-      CheckWriteOnlySecurity(CORSUsed, principal, hadCrossOriginRedirects);
+  bool writeOnly = CanvasUtils::CheckWriteOnlySecurity(CORSUsed, principal,
+                                                       hadCrossOriginRedirects);
 
   // Create ImageBitmap.
   RefPtr<layers::Image> data = aVideoEl.GetCurrentImage();
@@ -786,28 +953,36 @@ already_AddRefed<ImageBitmap> ImageBitmap::CreateInternal(
     aRv.Throw(NS_ERROR_NOT_AVAILABLE);
     return nullptr;
   }
-  RefPtr<ImageBitmap> ret = new ImageBitmap(aGlobal, data, writeOnly);
 
-  // Set the picture rectangle.
-  if (ret && aCropRect.isSome()) {
-    ret->SetPictureRect(aCropRect.ref(), aRv);
+  RefPtr<SourceSurface> surface = data->GetAsSourceSurface();
+  if (!surface) {
+    // preserve original behavior in case of unavailble surface
+    RefPtr<ImageBitmap> ret = new ImageBitmap(aGlobal, data, writeOnly);
+    return ret.forget();
   }
 
-  return ret.forget();
+  bool needToReportMemoryAllocation = false;
+
+  return CreateImageBitmapInternal(aGlobal, surface, aCropRect, aOptions,
+                                   writeOnly, needToReportMemoryAllocation,
+                                   false, gfxAlphaType::Premult, aRv);
 }
 
 /* static */
 already_AddRefed<ImageBitmap> ImageBitmap::CreateInternal(
     nsIGlobalObject* aGlobal, HTMLCanvasElement& aCanvasEl,
-    const Maybe<IntRect>& aCropRect, ErrorResult& aRv) {
+    const Maybe<IntRect>& aCropRect, const ImageBitmapOptions& aOptions,
+    ErrorResult& aRv) {
   if (aCanvasEl.Width() == 0 || aCanvasEl.Height() == 0) {
     aRv.Throw(NS_ERROR_DOM_INVALID_STATE_ERR);
     return nullptr;
   }
 
   bool writeOnly = true;
-  RefPtr<SourceSurface> surface =
-      GetSurfaceFromElement(aGlobal, aCanvasEl, &writeOnly, aRv);
+  gfxAlphaType alphaType = gfxAlphaType::NonPremult;
+
+  RefPtr<SourceSurface> surface = GetSurfaceFromElement(
+      aGlobal, aCanvasEl, &writeOnly, aOptions, &alphaType, aRv);
 
   if (NS_WARN_IF(aRv.Failed())) {
     return nullptr;
@@ -817,67 +992,43 @@ already_AddRefed<ImageBitmap> ImageBitmap::CreateInternal(
     writeOnly = aCanvasEl.IsWriteOnly();
   }
 
-  // Crop the source surface if needed.
-  RefPtr<SourceSurface> croppedSurface;
-  IntRect cropRect = aCropRect.valueOr(IntRect());
-
   // If the HTMLCanvasElement's rendering context is WebGL/WebGPU,
   // then the snapshot we got from the HTMLCanvasElement is
   // a DataSourceSurface which is a copy of the rendering context.
   // We handle cropping in this case.
   bool needToReportMemoryAllocation = false;
+  bool mustCopy = false;
+
   if ((aCanvasEl.GetCurrentContextType() == CanvasContextType::WebGL1 ||
        aCanvasEl.GetCurrentContextType() == CanvasContextType::WebGL2 ||
        aCanvasEl.GetCurrentContextType() == CanvasContextType::WebGPU) &&
       aCropRect.isSome()) {
-    RefPtr<DataSourceSurface> dataSurface = surface->GetDataSurface();
-    croppedSurface = CropAndCopyDataSourceSurface(dataSurface, cropRect);
-    cropRect.MoveTo(0, 0);
-    needToReportMemoryAllocation = true;
-  } else {
-    croppedSurface = surface;
+    mustCopy = true;
   }
 
-  if (NS_WARN_IF(!croppedSurface)) {
-    aRv.Throw(NS_ERROR_DOM_INVALID_STATE_ERR);
-    return nullptr;
-  }
-
-  // Create an Image from the SourceSurface.
-  RefPtr<layers::Image> data = CreateImageFromSurface(croppedSurface);
-
-  if (NS_WARN_IF(!data)) {
-    aRv.Throw(NS_ERROR_DOM_INVALID_STATE_ERR);
-    return nullptr;
-  }
-
-  RefPtr<ImageBitmap> ret = new ImageBitmap(aGlobal, data, writeOnly);
-
-  if (needToReportMemoryAllocation) {
-    ret->mAllocatedImageData = true;
-  }
-
-  // Set the picture rectangle.
-  if (ret && aCropRect.isSome()) {
-    ret->SetPictureRect(cropRect, aRv);
-  }
-
-  return ret.forget();
+  return CreateImageBitmapInternal(aGlobal, surface, aCropRect, aOptions,
+                                   writeOnly, needToReportMemoryAllocation,
+                                   mustCopy, alphaType, aRv);
 }
 
 /* static */
 already_AddRefed<ImageBitmap> ImageBitmap::CreateInternal(
     nsIGlobalObject* aGlobal, ImageData& aImageData,
-    const Maybe<IntRect>& aCropRect, ErrorResult& aRv) {
+    const Maybe<IntRect>& aCropRect, const ImageBitmapOptions& aOptions,
+    ErrorResult& aRv) {
   // Copy data into SourceSurface.
-  dom::Uint8ClampedArray array;
-  DebugOnly<bool> inited = array.Init(aImageData.GetDataObject());
-  MOZ_ASSERT(inited);
-
+  RootedSpiderMonkeyInterface<Uint8ClampedArray> array(RootingCx());
+  if (!array.Init(aImageData.GetDataObject())) {
+    aRv.ThrowInvalidStateError(
+        "Failed to extract Uint8ClampedArray from ImageData (security check "
+        "failed?)");
+    return nullptr;
+  }
   array.ComputeState();
   const SurfaceFormat FORMAT = SurfaceFormat::R8G8B8A8;
   // ImageData's underlying data is not alpha-premultiplied.
-  const auto alphaType = gfxAlphaType::NonPremult;
+  auto alphaType = gfxAlphaType::NonPremult;
+
   const uint32_t BYTES_PER_PIXEL = BytesPerPixel(FORMAT);
   const uint32_t imageWidth = aImageData.Width();
   const uint32_t imageHeight = aImageData.Height();
@@ -886,31 +1037,48 @@ already_AddRefed<ImageBitmap> ImageBitmap::CreateInternal(
   const gfx::IntSize imageSize(imageWidth, imageHeight);
 
   // Check the ImageData is neutered or not.
-  if (imageWidth == 0 || imageHeight == 0 ||
-      (imageWidth * imageHeight * BYTES_PER_PIXEL) != dataLength) {
-    aRv.Throw(NS_ERROR_DOM_INVALID_STATE_ERR);
+  if (imageWidth == 0 || imageHeight == 0) {
+    aRv.ThrowInvalidStateError("Passed-in image is empty");
+    return nullptr;
+  }
+
+  if ((imageWidth * imageHeight * BYTES_PER_PIXEL) != dataLength) {
+    aRv.ThrowInvalidStateError("Data size / image format mismatch");
     return nullptr;
   }
 
   // Create and Crop the raw data into a layers::Image
   RefPtr<layers::Image> data;
+
+  // If the data could move during a GC, copy it out into a local buffer that
+  // lives until a CreateImageFromRawData lower in the stack copies it.
+  // Reassure the static analysis that we know what we're doing.
+  size_t maxInline = JS_MaxMovableTypedArraySize();
+  uint8_t inlineDataBuffer[maxInline];
+  uint8_t* fixedData = array.FixedData(inlineDataBuffer, maxInline);
+
+  // Lie to the hazard analysis and say that we're done with everything that
+  // `array` was using (safe because the data buffer is fixed, and the holding
+  // JSObject is being kept alive elsewhere.)
+  array.Reset();
+
   if (NS_IsMainThread()) {
-    data = CreateImageFromRawData(imageSize, imageStride, FORMAT, array.Data(),
-                                  dataLength, aCropRect);
+    data = CreateImageFromRawData(imageSize, imageStride, FORMAT, fixedData,
+                                  dataLength, aCropRect, aOptions);
   } else {
     RefPtr<CreateImageFromRawDataInMainThreadSyncTask> task =
         new CreateImageFromRawDataInMainThreadSyncTask(
-            array.Data(), dataLength, imageStride, FORMAT, imageSize, aCropRect,
-            getter_AddRefs(data));
+            fixedData, dataLength, imageStride, FORMAT, imageSize, aCropRect,
+            getter_AddRefs(data), aOptions);
     task->Dispatch(Canceling, aRv);
   }
 
   if (NS_WARN_IF(!data)) {
-    aRv.Throw(NS_ERROR_DOM_INVALID_STATE_ERR);
+    aRv.ThrowInvalidStateError("Failed to create internal image");
     return nullptr;
   }
 
-  // Create an ImageBimtap.
+  // Create an ImageBitmap.
   RefPtr<ImageBitmap> ret =
       new ImageBitmap(aGlobal, data, false /* write-only */, alphaType);
 
@@ -925,7 +1093,8 @@ already_AddRefed<ImageBitmap> ImageBitmap::CreateInternal(
 /* static */
 already_AddRefed<ImageBitmap> ImageBitmap::CreateInternal(
     nsIGlobalObject* aGlobal, CanvasRenderingContext2D& aCanvasCtx,
-    const Maybe<IntRect>& aCropRect, ErrorResult& aRv) {
+    const Maybe<IntRect>& aCropRect, const ImageBitmapOptions& aOptions,
+    ErrorResult& aRv) {
   nsCOMPtr<nsPIDOMWindowInner> win = do_QueryInterface(aGlobal);
   nsGlobalWindowInner* window = nsGlobalWindowInner::Cast(win);
   if (NS_WARN_IF(!window) || !window->GetExtantDoc()) {
@@ -934,7 +1103,7 @@ already_AddRefed<ImageBitmap> ImageBitmap::CreateInternal(
   }
 
   window->GetExtantDoc()->WarnOnceAbout(
-      Document::eCreateImageBitmapCanvasRenderingContext2D);
+      DeprecatedOperations::eCreateImageBitmapCanvasRenderingContext2D);
 
   // Check write-only mode.
   bool writeOnly =
@@ -953,44 +1122,32 @@ already_AddRefed<ImageBitmap> ImageBitmap::CreateInternal(
     return nullptr;
   }
 
-  RefPtr<layers::Image> data = CreateImageFromSurface(surface);
+  bool needToReportMemoryAllocation = true;
 
-  if (NS_WARN_IF(!data)) {
-    aRv.Throw(NS_ERROR_NOT_AVAILABLE);
-    return nullptr;
-  }
-
-  RefPtr<ImageBitmap> ret = new ImageBitmap(aGlobal, data, writeOnly);
-
-  ret->mAllocatedImageData = true;
-
-  // Set the picture rectangle.
-  if (ret && aCropRect.isSome()) {
-    ret->SetPictureRect(aCropRect.ref(), aRv);
-  }
-
-  return ret.forget();
+  return CreateImageBitmapInternal(aGlobal, surface, aCropRect, aOptions,
+                                   writeOnly, needToReportMemoryAllocation,
+                                   false, gfxAlphaType::Premult, aRv);
 }
 
 /* static */
 already_AddRefed<ImageBitmap> ImageBitmap::CreateInternal(
     nsIGlobalObject* aGlobal, ImageBitmap& aImageBitmap,
-    const Maybe<IntRect>& aCropRect, ErrorResult& aRv) {
+    const Maybe<IntRect>& aCropRect, const ImageBitmapOptions& aOptions,
+    ErrorResult& aRv) {
   if (!aImageBitmap.mData) {
     aRv.Throw(NS_ERROR_DOM_INVALID_STATE_ERR);
     return nullptr;
   }
 
   RefPtr<layers::Image> data = aImageBitmap.mData;
-  RefPtr<ImageBitmap> ret = new ImageBitmap(
-      aGlobal, data, aImageBitmap.mWriteOnly, aImageBitmap.mAlphaType);
 
-  // Set the picture rectangle.
-  if (ret && aCropRect.isSome()) {
-    ret->SetPictureRect(aCropRect.ref(), aRv);
-  }
+  RefPtr<SourceSurface> surface = data->GetAsSourceSurface();
 
-  return ret.forget();
+  bool needToReportMemoryAllocation = false;
+
+  return CreateImageBitmapInternal(
+      aGlobal, surface, aCropRect, aOptions, aImageBitmap.mWriteOnly,
+      needToReportMemoryAllocation, false, aImageBitmap.mAlphaType, aRv);
 }
 
 class FulfillImageBitmapPromise {
@@ -1050,7 +1207,7 @@ static void AsyncFulfillImageBitmapPromise(Promise* aPromise,
 
 class CreateImageBitmapFromBlobRunnable;
 
-class CreateImageBitmapFromBlob final : public CancelableRunnable,
+class CreateImageBitmapFromBlob final : public DiscardableRunnable,
                                         public imgIContainerCallback,
                                         public nsIInputStreamCallback {
   friend class CreateImageBitmapFromBlobRunnable;
@@ -1062,7 +1219,8 @@ class CreateImageBitmapFromBlob final : public CancelableRunnable,
 
   static already_AddRefed<CreateImageBitmapFromBlob> Create(
       Promise* aPromise, nsIGlobalObject* aGlobal, Blob& aBlob,
-      const Maybe<IntRect>& aCropRect, nsIEventTarget* aMainThreadEventTarget);
+      const Maybe<IntRect>& aCropRect, nsIEventTarget* aMainThreadEventTarget,
+      const ImageBitmapOptions& aOptions);
 
   NS_IMETHOD Run() override {
     MOZ_ASSERT(IsCurrentThread());
@@ -1082,8 +1240,10 @@ class CreateImageBitmapFromBlob final : public CancelableRunnable,
   CreateImageBitmapFromBlob(Promise* aPromise, nsIGlobalObject* aGlobal,
                             already_AddRefed<nsIInputStream> aInputStream,
                             const Maybe<IntRect>& aCropRect,
-                            nsIEventTarget* aMainThreadEventTarget)
-      : CancelableRunnable("dom::CreateImageBitmapFromBlob"),
+                            nsIEventTarget* aMainThreadEventTarget,
+                            const ImageBitmapOptions& aOptions)
+      : DiscardableRunnable("dom::CreateImageBitmapFromBlob"),
+
         mMutex("dom::CreateImageBitmapFromBlob::mMutex"),
         mPromise(aPromise),
         mGlobalObject(aGlobal),
@@ -1091,6 +1251,7 @@ class CreateImageBitmapFromBlob final : public CancelableRunnable,
         mCropRect(aCropRect),
         mOriginalCropRect(aCropRect),
         mMainThreadEventTarget(aMainThreadEventTarget),
+        mOptions(aOptions),
         mThread(PR_GetCurrentThread()) {}
 
   virtual ~CreateImageBitmapFromBlob() = default;
@@ -1138,12 +1299,12 @@ class CreateImageBitmapFromBlob final : public CancelableRunnable,
   Maybe<IntRect> mCropRect;
   Maybe<IntRect> mOriginalCropRect;
   IntSize mSourceSize;
-
   nsCOMPtr<nsIEventTarget> mMainThreadEventTarget;
+  const ImageBitmapOptions mOptions;
   void* mThread;
 };
 
-NS_IMPL_ISUPPORTS_INHERITED(CreateImageBitmapFromBlob, CancelableRunnable,
+NS_IMPL_ISUPPORTS_INHERITED(CreateImageBitmapFromBlob, DiscardableRunnable,
                             imgIContainerCallback, nsIInputStreamCallback)
 
 class CreateImageBitmapFromBlobRunnable : public WorkerRunnable {
@@ -1171,7 +1332,8 @@ class CreateImageBitmapFromBlobRunnable : public WorkerRunnable {
 static void AsyncCreateImageBitmapFromBlob(Promise* aPromise,
                                            nsIGlobalObject* aGlobal,
                                            Blob& aBlob,
-                                           const Maybe<IntRect>& aCropRect) {
+                                           const Maybe<IntRect>& aCropRect,
+                                           const ImageBitmapOptions& aOptions) {
   // Let's identify the main-thread event target.
   nsCOMPtr<nsIEventTarget> mainThreadEventTarget;
   if (NS_IsMainThread()) {
@@ -1183,7 +1345,7 @@ static void AsyncCreateImageBitmapFromBlob(Promise* aPromise,
   }
 
   RefPtr<CreateImageBitmapFromBlob> task = CreateImageBitmapFromBlob::Create(
-      aPromise, aGlobal, aBlob, aCropRect, mainThreadEventTarget);
+      aPromise, aGlobal, aBlob, aCropRect, mainThreadEventTarget, aOptions);
   if (NS_WARN_IF(!task)) {
     aPromise->MaybeReject(NS_ERROR_DOM_INVALID_STATE_ERR);
     return;
@@ -1195,7 +1357,8 @@ static void AsyncCreateImageBitmapFromBlob(Promise* aPromise,
 /* static */
 already_AddRefed<Promise> ImageBitmap::Create(
     nsIGlobalObject* aGlobal, const ImageBitmapSource& aSrc,
-    const Maybe<gfx::IntRect>& aCropRect, ErrorResult& aRv) {
+    const Maybe<gfx::IntRect>& aCropRect, const ImageBitmapOptions& aOptions,
+    ErrorResult& aRv) {
   MOZ_ASSERT(aGlobal);
 
   RefPtr<Promise> promise = Promise::Create(aGlobal, aRv);
@@ -1224,41 +1387,41 @@ already_AddRefed<Promise> ImageBitmap::Create(
     MOZ_ASSERT(
         NS_IsMainThread(),
         "Creating ImageBitmap from HTMLImageElement off the main thread.");
-    imageBitmap =
-        CreateInternal(aGlobal, aSrc.GetAsHTMLImageElement(), aCropRect, aRv);
+    imageBitmap = CreateInternal(aGlobal, aSrc.GetAsHTMLImageElement(),
+                                 aCropRect, aOptions, aRv);
   } else if (aSrc.IsSVGImageElement()) {
     MOZ_ASSERT(
         NS_IsMainThread(),
         "Creating ImageBitmap from SVGImageElement off the main thread.");
-    imageBitmap =
-        CreateInternal(aGlobal, aSrc.GetAsSVGImageElement(), aCropRect, aRv);
+    imageBitmap = CreateInternal(aGlobal, aSrc.GetAsSVGImageElement(),
+                                 aCropRect, aOptions, aRv);
   } else if (aSrc.IsHTMLVideoElement()) {
     MOZ_ASSERT(
         NS_IsMainThread(),
         "Creating ImageBitmap from HTMLVideoElement off the main thread.");
-    imageBitmap =
-        CreateInternal(aGlobal, aSrc.GetAsHTMLVideoElement(), aCropRect, aRv);
+    imageBitmap = CreateInternal(aGlobal, aSrc.GetAsHTMLVideoElement(),
+                                 aCropRect, aOptions, aRv);
   } else if (aSrc.IsHTMLCanvasElement()) {
     MOZ_ASSERT(
         NS_IsMainThread(),
         "Creating ImageBitmap from HTMLCanvasElement off the main thread.");
-    imageBitmap =
-        CreateInternal(aGlobal, aSrc.GetAsHTMLCanvasElement(), aCropRect, aRv);
+    imageBitmap = CreateInternal(aGlobal, aSrc.GetAsHTMLCanvasElement(),
+                                 aCropRect, aOptions, aRv);
   } else if (aSrc.IsImageData()) {
-    imageBitmap =
-        CreateInternal(aGlobal, aSrc.GetAsImageData(), aCropRect, aRv);
+    imageBitmap = CreateInternal(aGlobal, aSrc.GetAsImageData(), aCropRect,
+                                 aOptions, aRv);
   } else if (aSrc.IsCanvasRenderingContext2D()) {
     MOZ_ASSERT(NS_IsMainThread(),
                "Creating ImageBitmap from CanvasRenderingContext2D off the "
                "main thread.");
     imageBitmap = CreateInternal(aGlobal, aSrc.GetAsCanvasRenderingContext2D(),
-                                 aCropRect, aRv);
+                                 aCropRect, aOptions, aRv);
   } else if (aSrc.IsImageBitmap()) {
-    imageBitmap =
-        CreateInternal(aGlobal, aSrc.GetAsImageBitmap(), aCropRect, aRv);
+    imageBitmap = CreateInternal(aGlobal, aSrc.GetAsImageBitmap(), aCropRect,
+                                 aOptions, aRv);
   } else if (aSrc.IsBlob()) {
     AsyncCreateImageBitmapFromBlob(promise, aGlobal, aSrc.GetAsBlob(),
-                                   aCropRect);
+                                   aCropRect, aOptions);
     return promise.forget();
   } else {
     MOZ_CRASH("Unsupported type!");
@@ -1415,7 +1578,8 @@ size_t BindingJSObjectMallocBytes(ImageBitmap* aBitmap) {
 /* static */
 already_AddRefed<CreateImageBitmapFromBlob> CreateImageBitmapFromBlob::Create(
     Promise* aPromise, nsIGlobalObject* aGlobal, Blob& aBlob,
-    const Maybe<IntRect>& aCropRect, nsIEventTarget* aMainThreadEventTarget) {
+    const Maybe<IntRect>& aCropRect, nsIEventTarget* aMainThreadEventTarget,
+    const ImageBitmapOptions& aOptions) {
   // Get the internal stream of the blob.
   nsCOMPtr<nsIInputStream> stream;
   ErrorResult error;
@@ -1436,7 +1600,8 @@ already_AddRefed<CreateImageBitmapFromBlob> CreateImageBitmapFromBlob::Create(
   }
 
   RefPtr<CreateImageBitmapFromBlob> task = new CreateImageBitmapFromBlob(
-      aPromise, aGlobal, stream.forget(), aCropRect, aMainThreadEventTarget);
+      aPromise, aGlobal, stream.forget(), aCropRect, aMainThreadEventTarget,
+      aOptions);
 
   // Nothing to do for the main-thread.
   if (NS_IsMainThread()) {
@@ -1570,10 +1735,14 @@ CreateImageBitmapFromBlob::OnImageReady(imgIContainer* aImgContainer,
   MOZ_ASSERT(aImgContainer);
 
   // Get the surface out.
-  uint32_t frameFlags = imgIContainer::FLAG_SYNC_DECODE |
-                        imgIContainer::FLAG_ASYNC_NOTIFY |
-                        imgIContainer::FLAG_WANT_DATA_SURFACE;
+  uint32_t frameFlags =
+      imgIContainer::FLAG_SYNC_DECODE | imgIContainer::FLAG_ASYNC_NOTIFY;
   uint32_t whichFrame = imgIContainer::FRAME_FIRST;
+
+  if (mOptions.mPremultiplyAlpha == PremultiplyAlpha::None) {
+    frameFlags |= imgIContainer::FLAG_DECODE_NO_PREMULTIPLY_ALPHA;
+  }
+
   RefPtr<SourceSurface> surface =
       aImgContainer->GetFrame(whichFrame, frameFlags);
 
@@ -1589,8 +1758,19 @@ CreateImageBitmapFromBlob::OnImageReady(imgIContainer* aImgContainer,
 
   // Crop the source surface if needed.
   RefPtr<SourceSurface> croppedSurface = surface;
+  RefPtr<DataSourceSurface> dataSurface = surface->GetDataSurface();
 
-  if (mCropRect.isSome()) {
+#ifdef DEBUG
+  // the returned dataSurface image memory is write protected in debug mode
+  // force a copy into unprotected memory as a side effect of
+  // CropAndCopyDataSourceSurface
+  bool copyRequired = mCropRect.isSome() ||
+                      mOptions.mImageOrientation == ImageOrientation::FlipY;
+#else
+  bool copyRequired = mCropRect.isSome();
+#endif
+
+  if (copyRequired) {
     // The blob is just decoded into a RasterImage and not optimized yet, so the
     // _surface_ we get is a DataSourceSurface which wraps the RasterImage's
     // raw buffer.
@@ -1603,9 +1783,26 @@ CreateImageBitmapFromBlob::OnImageReady(imgIContainer* aImgContainer,
     // TODO: Bug1189632 is going to refactor this create-from-blob part to
     //       decode the blob off the main thread. Re-check if we should do
     //       cropping at this moment again there.
-    RefPtr<DataSourceSurface> dataSurface = surface->GetDataSurface();
-    croppedSurface = CropAndCopyDataSourceSurface(dataSurface, mCropRect.ref());
-    mCropRect->MoveTo(0, 0);
+
+    IntRect cropRect =
+        mCropRect.isSome() ? mCropRect.ref() : dataSurface->GetRect();
+
+    croppedSurface = CropAndCopyDataSourceSurface(dataSurface, cropRect);
+    if (NS_WARN_IF(!croppedSurface)) {
+      MimeTypeAndDecodeAndCropBlobCompletedMainThread(
+          nullptr, NS_ERROR_DOM_INVALID_STATE_ERR);
+      return NS_OK;
+    }
+
+    dataSurface = croppedSurface->GetDataSurface();
+
+    if (mCropRect.isSome()) {
+      mCropRect->MoveTo(0, 0);
+    }
+  }
+
+  if (mOptions.mImageOrientation == ImageOrientation::FlipY) {
+    croppedSurface = FlipYDataSourceSurface(dataSurface);
   }
 
   if (NS_WARN_IF(!croppedSurface)) {
@@ -1673,9 +1870,15 @@ void CreateImageBitmapFromBlob::
     return;
   }
 
+  gfxAlphaType alphaType = gfxAlphaType::Premult;
+
+  if (mOptions.mPremultiplyAlpha == PremultiplyAlpha::None) {
+    alphaType = gfxAlphaType::NonPremult;
+  }
+
   // Create ImageBitmap object.
   RefPtr<ImageBitmap> imageBitmap =
-      new ImageBitmap(mGlobalObject, aImage, false /* write-only */);
+      new ImageBitmap(mGlobalObject, aImage, false /* write-only */, alphaType);
 
   if (mCropRect.isSome()) {
     ErrorResult rv;
@@ -1703,5 +1906,4 @@ void CreateImageBitmapFromBlob::WorkerShuttingDown() {
   mGlobalObject = nullptr;
 }
 
-}  // namespace dom
-}  // namespace mozilla
+}  // namespace mozilla::dom
