@@ -12,12 +12,14 @@
 #include <tuple>
 #include <vector>
 #include <unordered_map>
+#include "ipc/IPCMessageUtilsSpecializations.h"
 #include "mozilla/dom/QueueParamTraits.h"
 #include "mozilla/ipc/SharedMemoryBasic.h"
 #include "mozilla/Assertions.h"
 #include "mozilla/ipc/Shmem.h"
 #include "mozilla/ipc/ProtocolUtils.h"
 #include "mozilla/Logging.h"
+#include "mozilla/ScopeExit.h"
 #include "mozilla/TimeStamp.h"
 #include "mozilla/TypeTraits.h"
 #include "nsString.h"
@@ -610,7 +612,7 @@ namespace ipc {
 
 template <typename Actor>
 struct IPDLParamTraits<mozilla::dom::IpdlProducer<Actor>> {
-  typedef mozilla::dom::IpdlProducer<Actor> paramType;
+  using paramType = mozilla::dom::IpdlProducer<Actor>;
 
   static void Write(IPC::Message* aMsg, IProtocol* aActor,
                     const paramType& aParam) {
@@ -627,7 +629,7 @@ struct IPDLParamTraits<mozilla::dom::IpdlProducer<Actor>> {
 
 template <typename Actor>
 struct IPDLParamTraits<mozilla::dom::IpdlConsumer<Actor>> {
-  typedef mozilla::dom::IpdlConsumer<Actor> paramType;
+  using paramType = mozilla::dom::IpdlConsumer<Actor>;
 
   static void Write(IPC::Message* aMsg, IProtocol* aActor,
                     const paramType& aParam) {
@@ -646,7 +648,7 @@ struct IPDLParamTraits<mozilla::dom::IpdlConsumer<Actor>> {
 
 template <>
 struct IPDLParamTraits<mozilla::dom::IpdlQueueBuffer> {
-  typedef mozilla::dom::IpdlQueueBuffer paramType;
+  using paramType = mozilla::dom::IpdlQueueBuffer;
 
   static void Write(IPC::Message* aMsg, IProtocol* aActor,
                     const paramType& aParam) {
