@@ -8,6 +8,7 @@
 #ifndef _QUEUEPARAMTRAITS_H_
 #define _QUEUEPARAMTRAITS_H_ 1
 
+#include "ipc/EnumSerializer.h"
 #include "mozilla/gfx/2D.h"
 #include "mozilla/Assertions.h"
 #include "mozilla/IntegerRange.h"
@@ -17,6 +18,7 @@
 #include "mozilla/Logging.h"
 #include "mozilla/TimeStamp.h"
 #include "mozilla/TypeTraits.h"
+#include "nsExceptionHandler.h"
 #include "nsString.h"
 #include "WebGLTypes.h"
 
@@ -49,8 +51,8 @@ inline bool operator!(const QueueStatus status) { return !IsSuccess(status); }
 
 template <typename T>
 struct RemoveCVR {
-  typedef typename std::remove_reference<typename std::remove_cv<T>::type>::type
-      Type;
+  using Type =
+      typename std::remove_reference<typename std::remove_cv<T>::type>::type;
 };
 
 inline size_t UsedBytes(size_t aQueueBufferSize, size_t aRead, size_t aWrite) {
@@ -65,8 +67,8 @@ inline size_t FreeBytes(size_t aQueueBufferSize, size_t aRead, size_t aWrite) {
 
 template <typename T>
 struct IsTriviallySerializable
-    : public std::integral_constant<bool, std::is_enum<T>::value ||
-                                              std::is_arithmetic<T>::value> {};
+    : public std::integral_constant<bool, std::is_arithmetic<T>::value &&
+                                              !std::is_same<T, bool>::value> {};
 
 /**
  * QueueParamTraits provide the user with a way to implement PCQ argument
@@ -141,6 +143,13 @@ class Marshaller {
   }
 };
 
+template <typename T>
+inline Range<T> AsRange(T* const begin, T* const end) {
+  const auto size = MaybeAs<size_t>(end - begin);
+  MOZ_RELEASE_ASSERT(size);
+  return {begin, *size};
+}
+
 /**
  * Used to give QueueParamTraits a way to write to the Producer without
  * actually altering it, in case the transaction fails.
@@ -158,16 +167,22 @@ class ProducerView {
         mWrite(aWrite),
         mStatus(QueueStatus::kSuccess) {}
 
+  template <typename T>
+  QueueStatus WriteFromRange(const Range<const T>& src) {
+    if (!mStatus) return mStatus;
+    mProducer->WriteFromRange(src);
+    return mStatus;
+  }
   /**
    * Copy bytes from aBuffer to the producer if there is enough room.
    * aBufferSize must not be 0.
    */
-  inline QueueStatus Write(const uint8_t* begin, const uint8_t* end);
-
   template <typename T>
   inline QueueStatus Write(const T* begin, const T* end) {
-    return Write(reinterpret_cast<const uint8_t*>(begin),
-                 reinterpret_cast<const uint8_t*>(end));
+    MOZ_RELEASE_ASSERT(begin <= end);
+    if (!mStatus) return mStatus;
+    WriteFromRange(AsRange(begin, end));
+    return mStatus;
   }
 
   template <typename T>
@@ -214,12 +229,31 @@ class ConsumerView {
    * Read bytes from the consumer if there is enough data.  aBuffer may
    * be null (in which case the data is skipped)
    */
-  inline QueueStatus Read(uint8_t* begin, uint8_t* end);
-
   template <typename T>
-  inline QueueStatus Read(T* begin, T* end) {
-    return Read(reinterpret_cast<uint8_t*>(begin),
-                reinterpret_cast<uint8_t*>(end));
+  inline QueueStatus Read(T* const destBegin, T* const destEnd) {
+    MOZ_ASSERT(destBegin);
+    MOZ_RELEASE_ASSERT(destBegin <= destEnd);
+    if (!mStatus) return mStatus;
+
+    const auto dest = AsRange(destBegin, destEnd);
+    const auto view = ReadRange<T>(dest.length());
+    if (!view) return mStatus;
+    const auto byteSize = ByteSize(dest);
+    if (byteSize) {
+      memcpy(dest.begin().get(), view->begin().get(), byteSize);
+    }
+    return mStatus;
+  }
+
+  /// Return a view wrapping the shmem.
+  template <typename T>
+  inline Maybe<Range<const T>> ReadRange(const size_t elemCount) {
+    if (!mStatus) return {};
+    const auto view = mConsumer->template ReadRange<T>(elemCount);
+    if (!view) {
+      mStatus = QueueStatus::kTooSmall;
+    }
+    return view;
   }
 
   template <typename T>
@@ -249,29 +283,6 @@ class ConsumerView {
   QueueStatus mStatus;
 };
 
-template <typename T>
-QueueStatus ProducerView<T>::Write(const uint8_t* const begin,
-                                   const uint8_t* const end) {
-  MOZ_ASSERT(begin <= end);
-  if (!mStatus) return mStatus;
-  if (begin < end) {
-    MOZ_ASSERT(begin);
-    mStatus = mProducer->WriteObject(mRead, mWrite, begin, end - begin);
-  }
-  return mStatus;
-}
-
-template <typename T>
-QueueStatus ConsumerView<T>::Read(uint8_t* const begin, uint8_t* const end) {
-  MOZ_ASSERT(begin <= end);
-  if (!mStatus) return mStatus;
-  if (begin < end) {
-    MOZ_ASSERT(begin);
-    mStatus = mConsumer->ReadObject(mRead, mWrite, begin, end - begin);
-  }
-  return mStatus;
-}
-
 // ---------------------------------------------------------------
 
 /**
@@ -301,36 +312,62 @@ struct QueueParamTraits {
 
 // ---------------------------------------------------------------
 
+template <>
+struct QueueParamTraits<bool> {
+  using ParamType = bool;
+
+  template <typename U>
+  static QueueStatus Write(ProducerView<U>& aProducerView,
+                           const ParamType& aArg) {
+    uint8_t temp = aArg ? 1 : 0;
+    return aProducerView.WriteParam(temp);
+  }
+
+  template <typename U>
+  static QueueStatus Read(ConsumerView<U>& aConsumerView, ParamType* aArg) {
+    uint8_t temp;
+    if (IsSuccess(aConsumerView.ReadParam(&temp))) {
+      MOZ_ASSERT(temp == 1 || temp == 0);
+      *aArg = temp ? true : false;
+    }
+    return aConsumerView.GetStatus();
+  }
+};
+
+// ---------------------------------------------------------------
+
 // Adapted from IPC::EnumSerializer, this class safely handles enum values,
 // validating that they are in range using the same EnumValidators as IPDL
 // (namely ContiguousEnumValidator and ContiguousEnumValidatorInclusive).
 template <typename E, typename EnumValidator>
 struct EnumSerializer {
-  typedef E ParamType;
-  typedef typename std::underlying_type<E>::type DataType;
+  using ParamType = E;
+  using DataType = typename std::underlying_type<E>::type;
 
   template <typename U>
-  static auto Write(ProducerView<U>& aProducerView, const ParamType& aValue) {
-    MOZ_RELEASE_ASSERT(EnumValidator::IsLegalValue(aValue));
+  static QueueStatus Write(ProducerView<U>& aProducerView,
+                           const ParamType& aValue) {
+    MOZ_RELEASE_ASSERT(
+        EnumValidator::IsLegalValue(static_cast<DataType>(aValue)));
     return aProducerView.WriteParam(DataType(aValue));
   }
 
   template <typename U>
-  static bool Read(ConsumerView<U>& aConsumerView, ParamType* aResult) {
+  static QueueStatus Read(ConsumerView<U>& aConsumerView, ParamType* aResult) {
     DataType value;
     if (!aConsumerView.ReadParam(&value)) {
       CrashReporter::AnnotateCrashReport(
           CrashReporter::Annotation::IPCReadErrorReason, "Bad iter"_ns);
-      return IsSuccess(aConsumerView.GetStatus());
+      return aConsumerView.GetStatus();
     }
-    if (!EnumValidator::IsLegalValue(ParamType(value))) {
+    if (!EnumValidator::IsLegalValue(static_cast<DataType>(value))) {
       CrashReporter::AnnotateCrashReport(
           CrashReporter::Annotation::IPCReadErrorReason, "Illegal value"_ns);
-      return IsSuccess(aConsumerView.GetStatus());
+      return QueueStatus::kFatalError;
     }
 
     *aResult = ParamType(value);
-    return true;
+    return QueueStatus::kSuccess;
   }
 };
 
@@ -362,30 +399,33 @@ struct QueueParamTraits<webgl::TexUnpackBlobDesc> {
 
   template <typename U>
   static QueueStatus Write(ProducerView<U>& view, const ParamType& in) {
-    MOZ_ASSERT(!in.image);
-    const bool isSurf = bool(in.surf);
+    MOZ_RELEASE_ASSERT(!in.image);
+    const bool isDataSurf = bool(in.dataSurf);
     if (!view.WriteParam(in.imageTarget) || !view.WriteParam(in.size) ||
         !view.WriteParam(in.srcAlphaType) || !view.WriteParam(in.unpacking) ||
         !view.WriteParam(in.cpuData) || !view.WriteParam(in.pboOffset) ||
-        !view.WriteParam(isSurf)) {
+        !view.WriteParam(in.imageSize) || !view.WriteParam(in.sd) ||
+        !view.WriteParam(in.applyUnpackTransforms) ||
+        !view.WriteParam(isDataSurf)) {
       return view.GetStatus();
     }
-    if (isSurf) {
-      gfx::DataSourceSurface::ScopedMap map(in.surf,
-                                            gfx::DataSourceSurface::READ);
+    if (isDataSurf) {
+      const auto& surf = in.dataSurf;
+      gfx::DataSourceSurface::ScopedMap map(surf, gfx::DataSourceSurface::READ);
       if (!map.IsMapped()) {
         return QueueStatus::kOOMError;
       }
-      const auto& surfSize = in.surf->GetSize();
+      const auto& surfSize = surf->GetSize();
       const auto stride = *MaybeAs<size_t>(map.GetStride());
-      if (!view.WriteParam(surfSize) ||
-          !view.WriteParam(in.surf->GetFormat()) || !view.WriteParam(stride)) {
+      if (!view.WriteParam(surfSize) || !view.WriteParam(surf->GetFormat()) ||
+          !view.WriteParam(stride)) {
         return view.GetStatus();
       }
 
       const size_t dataSize = stride * surfSize.height;
       const auto& begin = map.GetData();
-      if (!view.Write(begin, begin + dataSize)) {
+      const auto range = Range<const uint8_t>{begin, dataSize};
+      if (!view.WriteFromRange(range)) {
         return view.GetStatus();
       }
     }
@@ -394,14 +434,17 @@ struct QueueParamTraits<webgl::TexUnpackBlobDesc> {
 
   template <typename U>
   static QueueStatus Read(ConsumerView<U>& view, ParamType* const out) {
-    bool isSurf;
+    bool isDataSurf;
     if (!view.ReadParam(&out->imageTarget) || !view.ReadParam(&out->size) ||
         !view.ReadParam(&out->srcAlphaType) ||
         !view.ReadParam(&out->unpacking) || !view.ReadParam(&out->cpuData) ||
-        !view.ReadParam(&out->pboOffset) || !view.ReadParam(&isSurf)) {
+        !view.ReadParam(&out->pboOffset) || !view.ReadParam(&out->imageSize) ||
+        !view.ReadParam(&out->sd) ||
+        !view.ReadParam(&out->applyUnpackTransforms) ||
+        !view.ReadParam(&isDataSurf)) {
       return view.GetStatus();
     }
-    if (isSurf) {
+    if (isDataSurf) {
       gfx::IntSize surfSize;
       gfx::SurfaceFormat format;
       size_t stride;
@@ -409,22 +452,15 @@ struct QueueParamTraits<webgl::TexUnpackBlobDesc> {
           !view.ReadParam(&stride)) {
         return view.GetStatus();
       }
-      out->surf = gfx::Factory::CreateDataSourceSurfaceWithStride(
-          surfSize, format, stride, true);
-      if (!out->surf) {
-        return QueueStatus::kOOMError;
-      }
-
-      gfx::DataSourceSurface::ScopedMap map(out->surf,
-                                            gfx::DataSourceSurface::WRITE);
-      if (!map.IsMapped()) {
-        return QueueStatus::kOOMError;
-      }
       const size_t dataSize = stride * surfSize.height;
-      const auto& begin = map.GetData();
-      if (!view.Read(begin, begin + dataSize)) {
-        return view.GetStatus();
-      }
+      const auto range = view.template ReadRange<uint8_t>(dataSize);
+      if (!range) return view.GetStatus();
+
+      // DataSourceSurface demands pointer-to-mutable.
+      const auto bytes = const_cast<uint8_t*>(range->begin().get());
+      out->dataSurf = gfx::Factory::CreateWrappingDataSourceSurface(
+          bytes, stride, surfSize, format);
+      MOZ_ASSERT(out->dataSurf);
     }
     return QueueStatus::kSuccess;
   }
@@ -799,11 +835,7 @@ struct QueueParamTraits<UniquePtr<T>> {
       return QueueStatus::kSuccess;
     }
 
-    T* obj = nullptr;
-    obj = new T();
-    if (!obj) {
-      return QueueStatus::kOOMError;
-    }
+    T* obj = new T();
     aArg->reset(obj);
     return aConsumerView.ReadParam(obj);
   }
