@@ -12,13 +12,19 @@
 #include <vector>
 
 #include "GLDefs.h"
+#include "ImageContainer.h"
 #include "mozilla/Casting.h"
 #include "mozilla/CheckedInt.h"
+#include "mozilla/MathAlgorithms.h"
 #include "mozilla/Range.h"
 #include "mozilla/RefCounted.h"
+#include "mozilla/ResultVariant.h"
+#include "mozilla/gfx/2D.h"
 #include "mozilla/gfx/BuildConstants.h"
 #include "mozilla/gfx/Point.h"
+#include "mozilla/gfx/Rect.h"
 #include "mozilla/ipc/Shmem.h"
+#include "mozilla/layers/LayersSurfaces.h"
 #include "gfxTypes.h"
 
 #include "nsTArray.h"
@@ -28,9 +34,9 @@
 
 // Manual reflection of WebIDL typedefs that are different from their
 // OpenGL counterparts.
-typedef int64_t WebGLsizeiptr;
-typedef int64_t WebGLintptr;
-typedef bool WebGLboolean;
+using WebGLsizeiptr = int64_t;
+using WebGLintptr = int64_t;
+using WebGLboolean = bool;
 
 // -
 
@@ -222,6 +228,7 @@ enum class WebGLExtensionID : uint8_t {
   EXT_texture_filter_anisotropic,
   EXT_texture_norm16,
   MOZ_debug,
+  OES_draw_buffers_indexed,
   OES_element_index_uint,
   OES_fbo_render_mipmap,
   OES_standard_derivatives,
@@ -293,6 +300,8 @@ class UniqueBuffer {
 namespace webgl {
 struct FormatUsageInfo;
 
+static constexpr GLenum kErrorPerfWarning = 0x10001;
+
 struct SampleableInfo final {
   const char* incompleteReason = nullptr;
   uint32_t levels = 0;
@@ -318,7 +327,7 @@ enum class UniformBaseType : uint8_t {
 };
 const char* ToString(UniformBaseType);
 
-typedef uint64_t ObjectId;
+using ObjectId = uint64_t;
 
 enum class BufferKind : uint8_t {
   Undefined,
@@ -378,8 +387,10 @@ struct WebGLContextOptions {
 
 // -
 
-template <typename T>
+template <typename _T>
 struct avec2 {
+  using T = _T;
+
   T x = T();
   T y = T();
 
@@ -405,10 +416,55 @@ struct avec2 {
 
   bool operator==(const avec2& rhs) const { return x == rhs.x && y == rhs.y; }
   bool operator!=(const avec2& rhs) const { return !(*this == rhs); }
+
+#define _(OP)                                 \
+  avec2 operator OP(const avec2& rhs) const { \
+    return {x OP rhs.x, y OP rhs.y};          \
+  }                                           \
+  avec2 operator OP(const T rhs) const { return {x OP rhs, y OP rhs}; }
+
+  _(+)
+  _(-)
+  _(*)
+  _(/)
+
+#undef _
+
+  avec2 Clamp(const avec2& min, const avec2& max) const {
+    return {mozilla::Clamp(x, min.x, max.x), mozilla::Clamp(y, min.y, max.y)};
+  }
+
+  // mozilla::Clamp doesn't work on floats, so be clear that this is a min+max
+  // helper.
+  avec2 ClampMinMax(const avec2& min, const avec2& max) const {
+    const auto ClampScalar = [](const T v, const T min, const T max) {
+      return std::max(min, std::min(v, max));
+    };
+    return {ClampScalar(x, min.x, max.x), ClampScalar(y, min.y, max.y)};
+  }
+
+  template <typename U>
+  U StaticCast() const {
+    return {static_cast<typename U::T>(x), static_cast<typename U::T>(y)};
+  }
 };
 
 template <typename T>
+avec2<T> MinExtents(const avec2<T>& a, const avec2<T>& b) {
+  return {std::min(a.x, b.x), std::min(a.y, b.y)};
+}
+
+template <typename T>
+avec2<T> MaxExtents(const avec2<T>& a, const avec2<T>& b) {
+  return {std::max(a.x, b.x), std::max(a.y, b.y)};
+}
+
+// -
+
+template <typename _T>
 struct avec3 {
+  using T = _T;
+
   T x = T();
   T y = T();
   T z = T();
@@ -436,10 +492,12 @@ struct avec3 {
   bool operator!=(const avec3& rhs) const { return !(*this == rhs); }
 };
 
-typedef avec2<int32_t> ivec2;
-typedef avec3<int32_t> ivec3;
-typedef avec2<uint32_t> uvec2;
-typedef avec3<uint32_t> uvec3;
+using ivec2 = avec2<int32_t>;
+using ivec3 = avec3<int32_t>;
+using uvec2 = avec2<uint32_t>;
+using uvec3 = avec3<uint32_t>;
+
+inline ivec2 AsVec(const gfx::IntSize& s) { return {s.width, s.height}; }
 
 // -
 
@@ -482,11 +540,16 @@ struct ReadPixelsDesc final {
   PixelPackState packState;
 };
 
-class ExtensionBits final {
+// -
+
+template <typename E>
+class EnumMask {
+ public:
   uint64_t mBits = 0;
 
+ private:
   struct BitRef final {
-    ExtensionBits& bits;
+    EnumMask& bits;
     const uint64_t mask;
 
     explicit operator bool() const { return bits.mBits & mask; }
@@ -501,15 +564,17 @@ class ExtensionBits final {
     }
   };
 
-  uint64_t Mask(const WebGLExtensionID i) const {
+  uint64_t Mask(const E i) const {
     return uint64_t{1} << static_cast<uint64_t>(i);
   }
 
  public:
-  BitRef operator[](const WebGLExtensionID i) { return {*this, Mask(i)}; }
-  bool operator[](const WebGLExtensionID i) const { return mBits & Mask(i); }
+  BitRef operator[](const E i) { return {*this, Mask(i)}; }
+  bool operator[](const E i) const { return mBits & Mask(i); }
 };
 
+class ExtensionBits : public EnumMask<WebGLExtensionID> {};
+
 // -
 
 enum class ContextLossReason : uint8_t {
@@ -537,6 +602,8 @@ struct InitContextDesc final {
   uint32_t principalKey = 0;
 };
 
+constexpr uint32_t kMaxTransformFeedbackSeparateAttribs = 4;
+
 struct Limits final {
   ExtensionBits supportedExtensions;
 
@@ -545,14 +612,13 @@ struct Limits final {
   uint32_t maxTex2dSize = 0;
   uint32_t maxTexCubeSize = 0;
   uint32_t maxVertexAttribs = 0;
-  std::array<uint32_t, 2> maxViewportDims = {};
+  uint32_t maxViewportDim = 0;
   std::array<float, 2> pointSizeRange = {{1, 1}};
   std::array<float, 2> lineWidthRange = {{1, 1}};
 
   // WebGL 2
   uint32_t maxTexArrayLayers = 0;
   uint32_t maxTex3dSize = 0;
-  uint32_t maxTransformFeedbackSeparateAttribs = 0;
   uint32_t maxUniformBufferBindings = 0;
   uint32_t uniformBufferOffsetAlignment = 0;
 
@@ -568,6 +634,7 @@ struct InitContextResult final {
   std::string error;
   WebGLContextOptions options;
   webgl::Limits limits;
+  EnumMask<layers::SurfaceDescriptor::Type> uploadableSdTypes;
 };
 
 // -
@@ -674,7 +741,7 @@ struct GetUniformData final {
 
 struct FrontBufferSnapshotIpc final {
   uvec2 surfSize = {};
-  mozilla::ipc::Shmem shmem = {};
+  Maybe<mozilla::ipc::Shmem> shmem = {};
 };
 
 struct ReadPixelsResult {
@@ -739,6 +806,11 @@ class RawBuffer final {
   const auto& begin() const { return mBegin; }
   const auto& size() const { return mLen; }
 
+  void Shrink(const size_t newLen) {
+    if (mLen <= newLen) return;
+    mLen = newLen;
+  }
+
   RawBuffer() = default;
 
   RawBuffer(const RawBuffer&) = delete;
@@ -911,9 +983,8 @@ struct WebGLPixelStore final {
   void Apply(gl::GLContext&, bool isWebgl2, const uvec3& uploadSize) const;
   bool AssertCurrent(gl::GLContext&, bool isWebgl2) const;
 
-  WebGLPixelStore ForUseWith(
-      const GLenum target, const uvec3& uploadSize,
-      const Maybe<gfx::IntSize>& structuredSrcSize) const {
+  WebGLPixelStore ForUseWith(const GLenum target, const uvec3& uploadSize,
+                             const Maybe<uvec2>& structuredSrcSize) const {
     auto ret = *this;
 
     if (!IsTexTarget3D(target)) {
@@ -922,7 +993,7 @@ struct WebGLPixelStore final {
     }
 
     if (structuredSrcSize) {
-      ret.mUnpackRowLength = structuredSrcSize->width;
+      ret.mUnpackRowLength = structuredSrcSize->x;
     }
 
     if (!ret.mUnpackRowLength) {
@@ -930,10 +1001,38 @@ struct WebGLPixelStore final {
     }
     if (!ret.mUnpackImageHeight) {
       ret.mUnpackImageHeight = uploadSize.y;
+
+      if (!IsTexTarget3D(target)) {
+        // 2D targets don't enforce skipRows+height<=imageHeight.
+        ret.mUnpackImageHeight += ret.mUnpackSkipRows;
+      }
     }
 
     return ret;
   }
+
+  CheckedInt<size_t> UsedPixelsPerRow(const uvec3& size) const {
+    if (!size.x || !size.y || !size.z) return 0;
+    return CheckedInt<size_t>(mUnpackSkipPixels) + size.x;
+  }
+
+  CheckedInt<size_t> FullRowsNeeded(const uvec3& size) const {
+    if (!size.x || !size.y || !size.z) return 0;
+
+    // The spec guarantees:
+    // * SKIP_PIXELS + width <= ROW_LENGTH.
+    // * SKIP_ROWS + height <= IMAGE_HEIGHT.
+    MOZ_ASSERT(mUnpackImageHeight);
+    auto skipFullRows =
+        CheckedInt<size_t>(mUnpackSkipImages) * mUnpackImageHeight;
+    skipFullRows += mUnpackSkipRows;
+
+    // Full rows in the final image, excluding the tail.
+    auto usedFullRows = CheckedInt<size_t>(size.z - 1) * mUnpackImageHeight;
+    usedFullRows += size.y - 1;
+
+    return skipFullRows + usedFullRows;
+  }
 };
 
 struct TexImageData final {
@@ -956,10 +1055,16 @@ struct TexUnpackBlobDesc final {
 
   Maybe<RawBuffer<>> cpuData;
   Maybe<uint64_t> pboOffset;
+
+  uvec2 imageSize;
   RefPtr<layers::Image> image;
-  RefPtr<gfx::DataSourceSurface> surf;
+  Maybe<layers::SurfaceDescriptor> sd;
+  RefPtr<gfx::DataSourceSurface> dataSurf;
 
   WebGLPixelStore unpacking;
+  bool applyUnpackTransforms = true;
+
+  void Shrink(const webgl::PackingInfo&);
 };
 
 }  // namespace webgl
@@ -990,6 +1095,24 @@ inline auto MakeRangeAbv(const T& abv)
   return {abv.Data(), abv.Length()};
 }
 
+// -
+
+constexpr auto kUniversalAlignment = alignof(std::max_align_t);
+
+template <typename T>
+inline size_t AlignmentOffset(const size_t alignment, const T posOrPtr) {
+  MOZ_ASSERT(alignment);
+  const auto begin = reinterpret_cast<uintptr_t>(posOrPtr);
+  const auto wholeMultiples = (begin + (alignment - 1)) / alignment;
+  const auto aligned = wholeMultiples * alignment;
+  return aligned - begin;
+}
+
+template <typename T>
+inline size_t ByteSize(const Range<T>& range) {
+  return range.length() * sizeof(T);
+}
+
 Maybe<Range<const uint8_t>> GetRangeFromView(const dom::ArrayBufferView& view,
                                              GLuint elemOffset,
                                              GLuint elemCountOverride);
@@ -1035,6 +1158,15 @@ inline void Memcpy(const RangedPtr<uint8_t>& destBytes,
 
 // -
 
+namespace webgl {
+
+// In theory, this number can be unbounded based on the driver. However, no
+// driver appears to expose more than 8. We might as well stop there too, for
+// now.
+// (http://opengl.gpuinfo.org/gl_stats_caps_single.php?listreportsbycap=GL_MAX_COLOR_ATTACHMENTS)
+inline constexpr size_t kMaxDrawBuffers = 8;
+}  // namespace webgl
+
 }  // namespace mozilla
 
 #endif
