@@ -14,6 +14,7 @@
 
 namespace mozilla {
 namespace dom {
+class ShadowRoot;
 
 class ElementInternals final : public nsISupports, public nsWrapperCache {
  public:
@@ -27,6 +28,8 @@ class ElementInternals final : public nsISupports, public nsWrapperCache {
   virtual JSObject* WrapObject(JSContext* aCx,
                                JS::Handle<JSObject*> aGivenProto) override;
 
+  ShadowRoot* GetShadowRoot() const;
+
  private:
   ~ElementInternals() = default;
 
