@@ -50,6 +50,7 @@ class DispatchChangeEventCallback;
 class File;
 class FileList;
 class FileSystemEntry;
+class FormData;
 class GetFilesHelper;
 class InputType;
 
@@ -117,6 +118,8 @@ class HTMLInputElement final : public TextControlElement,
   using nsGenericHTMLFormElementWithState::GetForm;
   using nsGenericHTMLFormElementWithState::GetFormAction;
   using nsIConstraintValidation::GetValidationMessage;
+  using ValueSetterOption = TextControlState::ValueSetterOption;
+  using ValueSetterOptions = TextControlState::ValueSetterOptions;
 
   enum class FromClone { no, yes };
 
@@ -143,18 +146,21 @@ class HTMLInputElement final : public TextControlElement,
   // Element
   virtual bool IsInteractiveHTMLContent() const override;
 
+  // nsGenericHTMLElement
+  virtual bool IsDisabledForEvents(WidgetEvent* aEvent) override;
+
+  // nsGenericHTMLFormElement
+  void SaveState() override;
+  MOZ_CAN_RUN_SCRIPT_BOUNDARY bool RestoreState(PresState* aState) override;
+
   // EventTarget
   virtual void AsyncEventRunning(AsyncEventDispatcher* aEvent) override;
 
   // Overriden nsIFormControl methods
   MOZ_CAN_RUN_SCRIPT_BOUNDARY
   NS_IMETHOD Reset() override;
-  NS_IMETHOD SubmitNamesValues(HTMLFormSubmission* aFormSubmission) override;
-  NS_IMETHOD SaveState() override;
-  MOZ_CAN_RUN_SCRIPT_BOUNDARY
-  virtual bool RestoreState(PresState* aState) override;
+  NS_IMETHOD SubmitNamesValues(FormData* aFormData) override;
   virtual bool AllowDrop() override;
-  virtual bool IsDisabledForEvents(WidgetEvent* aEvent) override;
 
   virtual void FieldSetDisabledChanged(bool aNotify) override;
 
@@ -198,7 +204,8 @@ class HTMLInputElement final : public TextControlElement,
 
   virtual EventStates IntrinsicState() const override;
 
- public:
+  void SetLastValueChangeWasInteractive(bool);
+
   // TextControlElement
   virtual nsresult SetValueChanged(bool aValueChanged) override;
   virtual bool IsSingleLineTextControl() const override;
@@ -222,13 +229,10 @@ class HTMLInputElement final : public TextControlElement,
   MOZ_CAN_RUN_SCRIPT virtual void UnbindFromFrame(
       nsTextControlFrame* aFrame) override;
   MOZ_CAN_RUN_SCRIPT virtual nsresult CreateEditor() override;
-  virtual void UpdateOverlayTextVisibility(bool aNotify) override;
   virtual void SetPreviewValue(const nsAString& aValue) override;
   virtual void GetPreviewValue(nsAString& aValue) override;
   virtual void EnablePreview() override;
   virtual bool IsPreviewEnabled() override;
-  virtual bool GetPlaceholderVisibility() override;
-  virtual bool GetPreviewVisibility() override;
   virtual void InitializeKeyboardEventListeners() override;
   virtual void OnValueChanged(ValueChangeKind) override;
   virtual void GetValueFromSetRangeText(nsAString& aValue) override;
@@ -324,7 +328,7 @@ class HTMLInputElement final : public TextControlElement,
   void MaybeUpdateAllValidityStates(bool aNotify) {
     // If you need to add new type which supports validationMessage, you should
     // add test cases into test_MozEditableElement_setUserInput.html.
-    if (mType == NS_FORM_INPUT_EMAIL) {
+    if (mType == FormControlType::InputEmail) {
       UpdateAllValidityStates(aNotify);
     }
   }
@@ -670,7 +674,7 @@ class HTMLInputElement final : public TextControlElement,
 
   already_AddRefed<nsINodeList> GetLabels();
 
-  void Select();
+  MOZ_CAN_RUN_SCRIPT void Select();
 
   Nullable<uint32_t> GetSelectionStart(ErrorResult& aRv);
   MOZ_CAN_RUN_SCRIPT void SetSelectionStart(const Nullable<uint32_t>& aValue,
@@ -914,15 +918,14 @@ class HTMLInputElement final : public TextControlElement,
    * @param aValue      String to set.
    * @param aOldValue   Previous value before setting aValue.
                         If previous value is unknown, aOldValue can be nullptr.
-   * @param aFlags      See TextControlState::SetValueFlags.
+   * @param aOptions    See TextControlState::ValueSetterOption.
    */
-  MOZ_CAN_RUN_SCRIPT
-  nsresult SetValueInternal(const nsAString& aValue, const nsAString* aOldValue,
-                            uint32_t aFlags);
-
-  MOZ_CAN_RUN_SCRIPT
-  nsresult SetValueInternal(const nsAString& aValue, uint32_t aFlags) {
-    return SetValueInternal(aValue, nullptr, aFlags);
+  MOZ_CAN_RUN_SCRIPT nsresult
+  SetValueInternal(const nsAString& aValue, const nsAString* aOldValue,
+                   const ValueSetterOptions& aOptions);
+  MOZ_CAN_RUN_SCRIPT nsresult SetValueInternal(
+      const nsAString& aValue, const ValueSetterOptions& aOptions) {
+    return SetValueInternal(aValue, nullptr, aOptions);
   }
 
   // Generic getter for the value that doesn't do experimental control type
@@ -965,9 +968,9 @@ class HTMLInputElement final : public TextControlElement,
   virtual void ResultForDialogSubmit(nsAString& aResult) override;
 
   /**
-   * Dispatch a select event. Returns true if the event was not cancelled.
+   * Dispatch a select event.
    */
-  bool DispatchSelectEvent(nsPresContext* aPresContext);
+  void DispatchSelectEvent(nsPresContext* aPresContext);
 
   void SelectAll(nsPresContext* aPresContext);
   bool IsImage() const {
@@ -979,7 +982,7 @@ class HTMLInputElement final : public TextControlElement,
    * Visit the group of radio buttons this radio belongs to
    * @param aVisitor the visitor to visit with
    */
-  nsresult VisitGroup(nsIRadioVisitor* aVisitor, bool aFlushContent);
+  nsresult VisitGroup(nsIRadioVisitor* aVisitor);
 
   /**
    * Do all the work that |SetChecked| does (radio button handling, etc.), but
@@ -1080,7 +1083,7 @@ class HTMLInputElement final : public TextControlElement,
    * Manages the internal data storage across type changes.
    */
   MOZ_CAN_RUN_SCRIPT
-  void HandleTypeChange(uint8_t aNewType, bool aNotify);
+  void HandleTypeChange(FormControlType aNewType, bool aNotify);
 
   enum class ForValueGetter { No, Yes };
 
@@ -1364,13 +1367,13 @@ class HTMLInputElement final : public TextControlElement,
   /**
    * Returns if the current type is an experimental mobile type.
    */
-  static bool IsExperimentalMobileType(uint8_t aType);
+  static bool IsExperimentalMobileType(FormControlType);
 
   /*
    * Returns if the current type is one of the date/time input types: date,
    * time, month, week and datetime-local.
    */
-  static bool IsDateTimeInputType(uint8_t aType);
+  static bool IsDateTimeInputType(FormControlType);
 
   /**
    * Returns whether getting `.value` as a string should sanitize the value.
@@ -1397,15 +1400,6 @@ class HTMLInputElement final : public TextControlElement,
   nsresult InitFilePicker(FilePickerType aType);
   nsresult InitColorPicker();
 
-  /**
-   * Use this function before trying to open a picker.
-   * It checks if the page is allowed to open a new pop-up.
-   * If it returns true, you should not create the picker.
-   *
-   * @return true if popup should be blocked, false otherwise
-   */
-  bool IsPopupBlocked() const;
-
   GetFilesHelper* GetOrCreateGetFilesHelper(bool aRecursiveFlag,
                                             ErrorResult& aRv);
 
@@ -1577,26 +1571,44 @@ class HTMLInputElement final : public TextControlElement,
    * Returns true if selection methods can be called on element
    */
   bool SupportsTextSelection() const {
-    return mType == NS_FORM_INPUT_TEXT || mType == NS_FORM_INPUT_SEARCH ||
-           mType == NS_FORM_INPUT_URL || mType == NS_FORM_INPUT_TEL ||
-           mType == NS_FORM_INPUT_PASSWORD;
+    switch (mType) {
+      case FormControlType::InputText:
+      case FormControlType::InputSearch:
+      case FormControlType::InputUrl:
+      case FormControlType::InputTel:
+      case FormControlType::InputPassword:
+        return true;
+      default:
+        return false;
+    }
   }
 
-  static bool MayFireChangeOnBlur(uint8_t aType) {
-    return IsSingleLineTextControl(false, aType) ||
-           aType == NS_FORM_INPUT_RANGE || aType == NS_FORM_INPUT_NUMBER ||
-           aType == NS_FORM_INPUT_TIME || aType == NS_FORM_INPUT_DATE;
+  static bool CreatesDateTimeWidget(FormControlType aType) {
+    return aType == FormControlType::InputDate ||
+           aType == FormControlType::InputTime ||
+           (aType == FormControlType::InputDatetimeLocal &&
+            StaticPrefs::dom_forms_datetime_local_widget());
   }
 
+  bool CreatesDateTimeWidget() const { return CreatesDateTimeWidget(mType); }
+
+  static bool MayFireChangeOnBlur(FormControlType aType) {
+    return IsSingleLineTextControl(false, aType) ||
+           CreatesDateTimeWidget(aType) ||
+           aType == FormControlType::InputRange ||
+           aType == FormControlType::InputNumber;
+  }
+
+  /**
+   * Fire an event when the password input field is removed from the DOM tree.
+   * This is now only used by the password manager.
+   */
+  void MaybeFireInputPasswordRemoved();
+
   /**
    * Checks if aDateTimeInputType should be supported.
    */
-  static bool IsDateTimeTypeSupported(uint8_t aDateTimeInputType);
-
-  static bool CreatesDateTimeWidget(uint8_t aType) {
-    return aType == NS_FORM_INPUT_TIME || aType == NS_FORM_INPUT_DATE;
-  }
-  bool CreatesDateTimeWidget() const { return CreatesDateTimeWidget(mType); }
+  static bool IsDateTimeTypeSupported(FormControlType);
 
   struct nsFilePickerFilter {
     nsFilePickerFilter() : mFilterMask(0) {}
@@ -1641,7 +1653,7 @@ class HTMLInputElement final : public TextControlElement,
 
    private:
     nsCOMPtr<nsIFilePicker> mFilePicker;
-    RefPtr<HTMLInputElement> mInput;
+    const RefPtr<HTMLInputElement> mInput;
   };
 };
 
