@@ -12,38 +12,69 @@
 #ifndef nsXULElement_h__
 #define nsXULElement_h__
 
+#include <stdint.h>
+#include <stdio.h>
+#include "ErrorList.h"
+#include "js/RootingAPI.h"
 #include "js/SourceText.h"
 #include "js/TracingAPI.h"
+#include "mozilla/AlreadyAddRefed.h"
+#include "mozilla/Assertions.h"
 #include "mozilla/Attributes.h"
-#include "nsAtom.h"
-#include "mozilla/dom/NodeInfo.h"
-#include "nsIControllers.h"
-#include "nsIURI.h"
-#include "nsLayoutCID.h"
-#include "AttrArray.h"
-#include "nsGkAtoms.h"
-#include "nsStringFwd.h"
-#include "nsStyledElement.h"
-#include "mozilla/dom/DOMRect.h"
-#include "mozilla/dom/Element.h"
+#include "mozilla/BasicEvents.h"
+#include "mozilla/RefPtr.h"
 #include "mozilla/dom/DOMString.h"
+#include "mozilla/dom/Element.h"
+#include "mozilla/dom/FragmentOrElement.h"
 #include "mozilla/dom/FromParser.h"
+#include "mozilla/dom/NameSpaceConstants.h"
+#include "mozilla/dom/NodeInfo.h"
+#include "nsAtom.h"
+#include "nsAttrName.h"
+#include "nsAttrValue.h"
+#include "nsCOMPtr.h"
+#include "nsCaseTreatment.h"
+#include "nsChangeHint.h"
+#include "nsCycleCollectionParticipant.h"
+#include "nsGkAtoms.h"
+#include "nsIContent.h"
+#include "nsINode.h"
+#include "nsISupports.h"
+#include "nsLiteralString.h"
+#include "nsString.h"
+#include "nsStyledElement.h"
+#include "nsTArray.h"
+#include "nsTLiteralString.h"
+#include "nscore.h"
 
-class nsXULPrototypeDocument;
-
+class JSObject;
+class JSScript;
+class nsAttrValueOrString;
+class nsIControllers;
 class nsIObjectInputStream;
 class nsIObjectOutputStream;
 class nsIOffThreadScriptReceiver;
+class nsIPrincipal;
+class nsIURI;
+class nsXULPrototypeDocument;
 class nsXULPrototypeNode;
-typedef nsTArray<RefPtr<nsXULPrototypeNode>> nsPrototypeArray;
+struct JSContext;
+
+using nsPrototypeArray = nsTArray<RefPtr<nsXULPrototypeNode>>;
+
+namespace JS {
+class CompileOptions;
+}
 
 namespace mozilla {
+class ErrorResult;
 class EventChainPreVisitor;
 class EventListenerManager;
 namespace css {
 class StyleRule;
 }  // namespace css
 namespace dom {
+class Document;
 class HTMLIFrameElement;
 class PrototypeDocumentContentSink;
 enum class CallerType : uint32_t;
@@ -185,6 +216,8 @@ class nsXULPrototypeScript : public nsXULPrototypeNode {
  private:
   virtual ~nsXULPrototypeScript();
 
+  void FillCompileOptions(JS::CompileOptions& options);
+
  public:
   virtual nsresult Serialize(
       nsIObjectOutputStream* aStream, nsXULPrototypeDocument* aProtoDoc,
@@ -296,7 +329,7 @@ ASSERT_NODE_FLAGS_SPACE(ELEMENT_TYPE_SPECIFIC_BITS_OFFSET + 2);
 
 class nsXULElement : public nsStyledElement {
  protected:
-  typedef mozilla::dom::Document Document;
+  using Document = mozilla::dom::Document;
 
   // Use Construct to construct elements instead of this constructor.
   explicit nsXULElement(already_AddRefed<mozilla::dom::NodeInfo>&& aNodeInfo);
@@ -331,7 +364,7 @@ class nsXULElement : public nsStyledElement {
   virtual void DestroyContent() override;
   virtual void DoneAddingChildren(bool aHaveNotified) override;
 
-#ifdef DEBUG
+#ifdef MOZ_DOM_LIST
   virtual void List(FILE* out, int32_t aIndent) const override;
   virtual void DumpContent(FILE* out, int32_t aIndent,
                            bool aDumpAll) const override {}
@@ -343,8 +376,9 @@ class nsXULElement : public nsStyledElement {
   MOZ_CAN_RUN_SCRIPT bool HasMenu();
   MOZ_CAN_RUN_SCRIPT void OpenMenu(bool aOpenFlag);
 
-  virtual bool PerformAccesskey(bool aKeyCausesActivation,
-                                bool aIsTrustedEvent) override;
+  MOZ_CAN_RUN_SCRIPT
+  virtual mozilla::Result<bool, nsresult> PerformAccesskey(
+      bool aKeyCausesActivation, bool aIsTrustedEvent) override;
   void ClickWithInputSource(uint16_t aInputSource, bool aIsTrustedEvent);
 
   virtual bool IsNodeOfType(uint32_t aFlags) const override;
@@ -362,7 +396,7 @@ class nsXULElement : public nsStyledElement {
 
   virtual bool IsEventAttributeNameInternal(nsAtom* aName) override;
 
-  typedef mozilla::dom::DOMString DOMString;
+  using DOMString = mozilla::dom::DOMString;
   void GetXULAttr(nsAtom* aName, DOMString& aResult) const {
     GetAttr(kNameSpaceID_None, aName, aResult);
   }
@@ -448,14 +482,6 @@ class nsXULElement : public nsStyledElement {
   void SetMaxHeight(const nsAString& aValue, mozilla::ErrorResult& rv) {
     SetXULAttr(nsGkAtoms::maxheight, aValue, rv);
   }
-  void GetLeft(DOMString& aValue) const { GetXULAttr(nsGkAtoms::left, aValue); }
-  void SetLeft(const nsAString& aValue, mozilla::ErrorResult& rv) {
-    SetXULAttr(nsGkAtoms::left, aValue, rv);
-  }
-  void GetTop(DOMString& aValue) const { GetXULAttr(nsGkAtoms::top, aValue); }
-  void SetTop(const nsAString& aValue, mozilla::ErrorResult& rv) {
-    SetXULAttr(nsGkAtoms::top, aValue, rv);
-  }
   void GetTooltipText(DOMString& aValue) const {
     GetXULAttr(nsGkAtoms::tooltiptext, aValue);
   }
@@ -466,10 +492,6 @@ class nsXULElement : public nsStyledElement {
   void SetSrc(const nsAString& aValue, mozilla::ErrorResult& rv) {
     SetXULAttr(nsGkAtoms::src, aValue, rv);
   }
-  bool AllowEvents() const { return BoolAttrIsTrue(nsGkAtoms::allowevents); }
-  void SetAllowEvents(bool aAllowEvents) {
-    SetXULBoolAttr(nsGkAtoms::allowevents, aAllowEvents);
-  }
   nsIControllers* GetControllers(mozilla::ErrorResult& rv);
   void Click(mozilla::dom::CallerType aCallerType);
   MOZ_CAN_RUN_SCRIPT_BOUNDARY void DoCommand();
@@ -483,8 +505,6 @@ class nsXULElement : public nsStyledElement {
 
   bool IsInteractiveHTMLContent() const override;
 
-  void MaybeUpdatePrivateLifetime();
-
  protected:
   ~nsXULElement();
 
@@ -537,7 +557,8 @@ class nsXULElement : public nsStyledElement {
     return slots ? slots->mControllers.get() : nullptr;
   }
 
-  void UnregisterAccessKey(const nsAString& aOldValue);
+  bool SupportsAccessKey() const;
+  void RegUnRegAccessKey(bool aDoReg) override;
   bool BoolAttrIsTrue(nsAtom* aName) const;
 
   friend nsXULElement* NS_NewBasicXULElement(
