@@ -20,30 +20,35 @@
  *   commands.json, update it, and write it back out.
  */
 
-this.EXPORTED_SYMBOLS = [
-  "ClientEngine",
-  "ClientsRec"
-];
+var EXPORTED_SYMBOLS = ["ClientEngine", "ClientsRec"];
 
-var {classes: Cc, interfaces: Ci, utils: Cu} = Components;
+const { Services } = ChromeUtils.import("resource://gre/modules/Services.jsm");
+const { Async } = ChromeUtils.import("resource://services-common/async.js");
+const {
+  DEVICE_TYPE_DESKTOP,
+  DEVICE_TYPE_MOBILE,
+  SCORE_INCREMENT_XLARGE,
+  SINGLE_USER_THRESHOLD,
+  SYNC_API_VERSION,
+} = ChromeUtils.import("resource://services-sync/constants.js");
+const { Store, SyncEngine, LegacyTracker } = ChromeUtils.import(
+  "resource://services-sync/engines.js"
+);
+const { CryptoWrapper } = ChromeUtils.import(
+  "resource://services-sync/record.js"
+);
+const { Resource } = ChromeUtils.import("resource://services-sync/resource.js");
+const { Svc, Utils } = ChromeUtils.import("resource://services-sync/util.js");
 
-Cu.import("resource://gre/modules/XPCOMUtils.jsm");
-Cu.import("resource://gre/modules/Services.jsm");
-Cu.import("resource://services-common/async.js");
-Cu.import("resource://services-sync/constants.js");
-Cu.import("resource://services-sync/engines.js");
-Cu.import("resource://services-sync/record.js");
-Cu.import("resource://services-sync/resource.js");
-Cu.import("resource://services-sync/util.js");
+ChromeUtils.defineModuleGetter(
+  this,
+  "fxAccounts",
+  "resource://gre/modules/FxAccounts.jsm"
+);
 
-XPCOMUtils.defineLazyModuleGetter(this, "fxAccounts",
-  "resource://gre/modules/FxAccounts.jsm");
-
-XPCOMUtils.defineLazyModuleGetter(this, "getRepairRequestor",
-  "resource://services-sync/collection_repair.js");
-
-XPCOMUtils.defineLazyModuleGetter(this, "getRepairResponder",
-  "resource://services-sync/collection_repair.js");
+const { PREF_ACCOUNT_ROOT } = ChromeUtils.import(
+  "resource://gre/modules/FxAccountsCommon.js"
+);
 
 const CLIENTS_TTL = 1814400; // 21 days
 const CLIENTS_TTL_REFRESH = 604800; // 7 days
@@ -52,47 +57,58 @@ const STALE_CLIENT_REMOTE_AGE = 604800; // 7 days
 // TTL of the message sent to another device when sending a tab
 const NOTIFY_TAB_SENT_TTL_SECS = 1 * 3600; // 1 hour
 
+// How often we force a refresh of the FxA device list.
+const REFRESH_FXA_DEVICE_INTERVAL_MS = 2 * 60 * 60 * 1000; // 2 hours
+
 // Reasons behind sending collection_changed push notifications.
 const COLLECTION_MODIFIED_REASON_SENDTAB = "sendtab";
 const COLLECTION_MODIFIED_REASON_FIRSTSYNC = "firstsync";
 
 const SUPPORTED_PROTOCOL_VERSIONS = [SYNC_API_VERSION];
-const LAST_MODIFIED_ON_PROCESS_COMMAND_PREF = "services.sync.clients.lastModifiedOnProcessCommands";
+const LAST_MODIFIED_ON_PROCESS_COMMAND_PREF =
+  "services.sync.clients.lastModifiedOnProcessCommands";
 
 function hasDupeCommand(commands, action) {
   if (!commands) {
     return false;
   }
-  return commands.some(other => other.command == action.command &&
-    Utils.deepEquals(other.args, action.args));
+  return commands.some(
+    other =>
+      other.command == action.command &&
+      Utils.deepEquals(other.args, action.args)
+  );
 }
 
-this.ClientsRec = function ClientsRec(collection, id) {
+function ClientsRec(collection, id) {
   CryptoWrapper.call(this, collection, id);
-};
+}
 ClientsRec.prototype = {
   __proto__: CryptoWrapper.prototype,
   _logName: "Sync.Record.Clients",
-  ttl: CLIENTS_TTL
+  ttl: CLIENTS_TTL,
 };
 
-Utils.deferGetSet(ClientsRec,
-                  "cleartext",
-                  ["name", "type", "commands",
-                   "version", "protocols",
-                   "formfactor", "os", "appPackage", "application", "device",
-                   "fxaDeviceId"]);
+Utils.deferGetSet(ClientsRec, "cleartext", [
+  "name",
+  "type",
+  "commands",
+  "version",
+  "protocols",
+  "formfactor",
+  "os",
+  "appPackage",
+  "application",
+  "device",
+  "fxaDeviceId",
+]);
 
-
-this.ClientEngine = function ClientEngine(service) {
+function ClientEngine(service) {
   SyncEngine.call(this, "Clients", service);
 
-  // Reset the last sync timestamp on every startup so that we fetch all clients
-  this.resetLastSync();
   this.fxAccounts = fxAccounts;
-  this.addClientCommandQueue = Promise.resolve();
+  this.addClientCommandQueue = Async.asyncQueueCaller(this._log);
   Utils.defineLazyIDProperty(this, "localID", "services.sync.client.GUID");
-};
+}
 ClientEngine.prototype = {
   __proto__: SyncEngine.prototype,
   _storeObj: ClientStore,
@@ -101,6 +117,12 @@ ClientEngine.prototype = {
   allowSkippedRecord: false,
   _knownStaleFxADeviceIds: null,
   _lastDeviceCounts: null,
+  _lastFxaDeviceRefresh: 0,
+
+  async initialize() {
+    // Reset the last sync timestamp on every startup so that we fetch all clients
+    await this.resetLastSync();
+  },
 
   // These two properties allow us to avoid replaying the same commands
   // continuously if we cannot manage to upload our own record.
@@ -113,6 +135,10 @@ ClientEngine.prototype = {
     Services.prefs.setIntPref(LAST_MODIFIED_ON_PROCESS_COMMAND_PREF, value);
   },
 
+  get isFirstSync() {
+    return !this.lastRecordUpload;
+  },
+
   // Always sync client data as it controls other sync behavior
   get enabled() {
     return true;
@@ -148,7 +174,7 @@ ClientEngine.prototype = {
     };
 
     for (let id in this._store._remoteClients) {
-      let {name, type, stale} = this._store._remoteClients[id];
+      let { name, type, stale } = this._store._remoteClients[id];
       if (!stale) {
         stats.hasMobile = stats.hasMobile || type == DEVICE_TYPE_MOBILE;
         stats.names.push(name);
@@ -190,26 +216,20 @@ ClientEngine.prototype = {
 
   get brandName() {
     let brand = Services.strings.createBundle(
-      "chrome://branding/locale/brand.properties");
+      "chrome://branding/locale/brand.properties"
+    );
     return brand.GetStringFromName("brandShortName");
   },
 
   get localName() {
-    return Utils.getDeviceName();
+    return this.fxAccounts.device.getLocalName();
   },
   set localName(value) {
-    Svc.Prefs.set("client.name", value);
-    // Update the registration in the background.
-    this.fxAccounts.updateDeviceRegistration().catch(error => {
-      this._log.warn("failed to update fxa device registration", error);
-    });
+    this.fxAccounts.device.setLocalName(value);
   },
 
   get localType() {
-    return Utils.getDeviceType();
-  },
-  set localType(value) {
-    Svc.Prefs.set("client.type", value);
+    return this.fxAccounts.device.getLocalType();
   },
 
   getClientName(id) {
@@ -227,10 +247,28 @@ ClientEngine.prototype = {
     return null;
   },
 
-  isMobile: function isMobile(id) {
-    if (this._store._remoteClients[id])
-      return this._store._remoteClients[id].type == DEVICE_TYPE_MOBILE;
-    return false;
+  getClientByFxaDeviceId(fxaDeviceId) {
+    for (let id in this._store._remoteClients) {
+      let client = this._store._remoteClients[id];
+      if (client.stale) {
+        continue;
+      }
+      if (client.fxaDeviceId == fxaDeviceId) {
+        return client;
+      }
+    }
+    return null;
+  },
+
+  getClientType(id) {
+    const client = this._store._remoteClients[id];
+    if (client.type == DEVICE_TYPE_DESKTOP) {
+      return "desktop";
+    }
+    if (client.formfactor && client.formfactor.includes("tablet")) {
+      return "tablet";
+    }
+    return "phone";
   },
 
   async _readCommands() {
@@ -283,8 +321,7 @@ ClientEngine.prototype = {
   },
 
   async _addClientCommand(clientId, command) {
-    return this.addClientCommandQueue = (async () => {
-      await this.addClientCommandQueue;
+    this.addClientCommandQueue.enqueueCall(async () => {
       try {
         const localCommands = await this._readCommands();
         const localClientCommands = localCommands[clientId] || [];
@@ -305,7 +342,9 @@ ClientEngine.prototype = {
         this._log.error(e);
         return false;
       }
-    })();
+    });
+
+    return this.addClientCommandQueue.promiseCallsComplete();
   },
 
   async _removeClientCommands(clientId) {
@@ -316,12 +355,14 @@ ClientEngine.prototype = {
 
   async updateKnownStaleClients() {
     this._log.debug("Updating the known stale clients");
-    await this._refreshKnownStaleClients();
-    let localFxADeviceId = await fxAccounts.getDeviceId();
+    // _fetchFxADevices side effect updates this._knownStaleFxADeviceIds.
+    await this._fetchFxADevices();
+    let localFxADeviceId = await fxAccounts.device.getLocalId();
     // Process newer records first, so that if we hit a record with a device ID
     // we've seen before, we can mark it stale immediately.
-    let clientList = Object.values(this._store._remoteClients).sort((a, b) =>
-      b.serverLastModified - a.serverLastModified);
+    let clientList = Object.values(this._store._remoteClients).sort(
+      (a, b) => b.serverLastModified - a.serverLastModified
+    );
     let seenDeviceIds = new Set([localFxADeviceId]);
     for (let client of clientList) {
       // Clients might not have an `fxaDeviceId` if they fail the FxA
@@ -330,11 +371,15 @@ ClientEngine.prototype = {
         continue;
       }
       if (this._knownStaleFxADeviceIds.includes(client.fxaDeviceId)) {
-        this._log.info(`Hiding stale client ${client.id} - in known stale clients list`);
+        this._log.info(
+          `Hiding stale client ${client.id} - in known stale clients list`
+        );
         client.stale = true;
       } else if (seenDeviceIds.has(client.fxaDeviceId)) {
-        this._log.info(`Hiding stale client ${client.id}` +
-                       ` - duplicate device id ${client.fxaDeviceId}`);
+        this._log.info(
+          `Hiding stale client ${client.id}` +
+            ` - duplicate device id ${client.fxaDeviceId}`
+        );
         client.stale = true;
       } else {
         seenDeviceIds.add(client.fxaDeviceId);
@@ -342,46 +387,50 @@ ClientEngine.prototype = {
     }
   },
 
-  // We assume that clients not present in the FxA Device Manager list have been
-  // disconnected and so are stale
-  async _refreshKnownStaleClients() {
+  async _fetchFxADevices() {
+    // We only force a refresh periodically to keep the load on the servers
+    // down, and because we expect FxA to have received a push message in
+    // most cases when the FxA device list would have changed. For this reason
+    // we still go ahead and check the stale list even if we didn't force a
+    // refresh.
+    let now = this.fxAccounts._internal.now(); // tests mock this .now() impl.
+    if (now - REFRESH_FXA_DEVICE_INTERVAL_MS > this._lastFxaDeviceRefresh) {
+      this._lastFxaDeviceRefresh = now;
+      try {
+        await this.fxAccounts.device.refreshDeviceList();
+      } catch (e) {
+        this._log.error("Could not refresh the FxA device list", e);
+      }
+    }
+
+    // We assume that clients not present in the FxA Device Manager list have been
+    // disconnected and so are stale
     this._log.debug("Refreshing the known stale clients list");
     let localClients = Object.values(this._store._remoteClients)
-                             .filter(client => client.fxaDeviceId) // iOS client records don't have fxaDeviceId
-                             .map(client => client.fxaDeviceId);
-    let fxaClients;
-    try {
-      let deviceList = await this.fxAccounts.getDeviceList();
-      fxaClients = deviceList.map(device => device.id);
-    } catch (ex) {
-      this._log.error("Could not retrieve the FxA device list", ex);
-      this._knownStaleFxADeviceIds = [];
-      return;
-    }
+      .filter(client => client.fxaDeviceId) // iOS client records don't have fxaDeviceId
+      .map(client => client.fxaDeviceId);
+    const fxaClients = this.fxAccounts.device.recentDeviceList
+      ? this.fxAccounts.device.recentDeviceList.map(device => device.id)
+      : [];
     this._knownStaleFxADeviceIds = Utils.arraySub(localClients, fxaClients);
   },
 
   async _syncStartup() {
-    this.isFirstSync = !this.lastRecordUpload;
     // Reupload new client record periodically.
     if (Date.now() / 1000 - this.lastRecordUpload > CLIENTS_TTL_REFRESH) {
-      this._tracker.addChangedID(this.localID);
-      this.lastRecordUpload = Date.now() / 1000;
+      await this._tracker.addChangedID(this.localID);
     }
     return SyncEngine.prototype._syncStartup.call(this);
   },
 
   async _processIncoming() {
     // Fetch all records from the server.
-    this.lastSync = 0;
+    await this.resetLastSync();
     this._incomingClients = {};
     try {
       await SyncEngine.prototype._processIncoming.call(this);
-      // Refresh the known stale clients list at startup and when we receive
-      // "device connected/disconnected" push notifications.
-      if (!this._knownStaleFxADeviceIds) {
-        await this._refreshKnownStaleClients();
-      }
+      // Update FxA Device list.
+      await this._fetchFxADevices();
       // Since clients are synced unconditionally, any records in the local store
       // that don't exist on the server must be for disconnected clients. Remove
       // them, so that we don't upload records with commands for clients that will
@@ -393,24 +442,32 @@ ClientEngine.prototype = {
           await this._removeRemoteClient(id);
         }
       }
-      let localFxADeviceId = await fxAccounts.getDeviceId();
+      let localFxADeviceId = await fxAccounts.device.getLocalId();
       // Bug 1264498: Mobile clients don't remove themselves from the clients
       // collection when the user disconnects Sync, so we mark as stale clients
       // with the same name that haven't synced in over a week.
       // (Note we can't simply delete them, or we re-apply them next sync - see
       // bug 1287687)
-      this._localClientLastModified = Math.round(this._incomingClients[this.localID]);
+      this._localClientLastModified = Math.round(
+        this._incomingClients[this.localID]
+      );
       delete this._incomingClients[this.localID];
       let names = new Set([this.localName]);
       let seenDeviceIds = new Set([localFxADeviceId]);
-      let idToLastModifiedList = Object.entries(this._incomingClients)
-                                 .sort((a, b) => b[1] - a[1]);
+      let idToLastModifiedList = Object.entries(this._incomingClients).sort(
+        (a, b) => b[1] - a[1]
+      );
       for (let [id, serverLastModified] of idToLastModifiedList) {
         let record = this._store._remoteClients[id];
         // stash the server last-modified time on the record.
         record.serverLastModified = serverLastModified;
-        if (record.fxaDeviceId && this._knownStaleFxADeviceIds.includes(record.fxaDeviceId)) {
-          this._log.info(`Hiding stale client ${id} - in known stale clients list`);
+        if (
+          record.fxaDeviceId &&
+          this._knownStaleFxADeviceIds.includes(record.fxaDeviceId)
+        ) {
+          this._log.info(
+            `Hiding stale client ${id} - in known stale clients list`
+          );
           record.stale = true;
         }
         if (!names.has(record.name)) {
@@ -427,8 +484,10 @@ ClientEngine.prototype = {
           continue;
         }
         if (record.fxaDeviceId && seenDeviceIds.has(record.fxaDeviceId)) {
-          this._log.info(`Hiding stale client ${record.id}` +
-                         ` - duplicate device id ${record.fxaDeviceId}`);
+          this._log.info(
+            `Hiding stale client ${record.id}` +
+              ` - duplicate device id ${record.fxaDeviceId}`
+          );
           record.stale = true;
         } else if (record.fxaDeviceId) {
           seenDeviceIds.add(record.fxaDeviceId);
@@ -441,7 +500,9 @@ ClientEngine.prototype = {
 
   async _uploadOutgoing() {
     this._currentlySyncingCommands = await this._prepareCommandsForUpload();
-    const clientWithPendingCommands = Object.keys(this._currentlySyncingCommands);
+    const clientWithPendingCommands = Object.keys(
+      this._currentlySyncingCommands
+    );
     for (let clientId of clientWithPendingCommands) {
       if (this._store._remoteClients[clientId] || this.localID == clientId) {
         this._modified.set(clientId, 0);
@@ -450,9 +511,12 @@ ClientEngine.prototype = {
     let updatedIDs = this._modified.ids();
     await SyncEngine.prototype._uploadOutgoing.call(this);
     // Record the response time as the server time for each item we uploaded.
+    let lastSync = await this.getLastSync();
     for (let id of updatedIDs) {
-      if (id != this.localID) {
-        this._store._remoteClients[id].serverLastModified = this.lastSync;
+      if (id == this.localID) {
+        this.lastRecordUpload = lastSync;
+      } else {
+        this._store._remoteClients[id].serverLastModified = lastSync;
       }
     }
   },
@@ -464,17 +528,23 @@ ClientEngine.prototype = {
       const commandChanges = this._currentlySyncingCommands[id];
       if (id == this.localID) {
         if (this.isFirstSync) {
-          this._log.info("Uploaded our client record for the first time, notifying other clients.");
+          this._log.info(
+            "Uploaded our client record for the first time, notifying other clients."
+          );
           this._notifyClientRecordUploaded();
         }
         if (this.localCommands) {
-          this.localCommands = this.localCommands.filter(command => !hasDupeCommand(commandChanges, command));
+          this.localCommands = this.localCommands.filter(
+            command => !hasDupeCommand(commandChanges, command)
+          );
         }
       } else {
         const clientRecord = this._store._remoteClients[id];
         if (!commandChanges || !clientRecord) {
           // should be impossible, else we wouldn't have been writing it.
-          this._log.warn("No command/No record changes for a client we uploaded");
+          this._log.warn(
+            "No command/No record changes for a client we uploaded"
+          );
           continue;
         }
         // fixup the client record, so our copy of _remoteClients matches what we uploaded.
@@ -510,13 +580,20 @@ ClientEngine.prototype = {
 
   _notifyOtherClientsModified(ids) {
     // We are not waiting on this promise on purpose.
-    this._notifyCollectionChanged(ids, NOTIFY_TAB_SENT_TTL_SECS,
-                                  COLLECTION_MODIFIED_REASON_SENDTAB);
+    this._notifyCollectionChanged(
+      ids,
+      NOTIFY_TAB_SENT_TTL_SECS,
+      COLLECTION_MODIFIED_REASON_SENDTAB
+    );
   },
 
   _notifyClientRecordUploaded() {
     // We are not waiting on this promise on purpose.
-    this._notifyCollectionChanged(null, 0, COLLECTION_MODIFIED_REASON_FIRSTSYNC);
+    this._notifyCollectionChanged(
+      null,
+      0,
+      COLLECTION_MODIFIED_REASON_FIRSTSYNC
+    );
   },
 
   /**
@@ -530,12 +607,12 @@ ClientEngine.prototype = {
       command: "sync:collection_changed",
       data: {
         collections: ["clients"],
-        reason
-      }
+        reason,
+      },
     };
     let excludedIds = null;
     if (!ids) {
-      const localFxADeviceId = await fxAccounts.getDeviceId();
+      const localFxADeviceId = await fxAccounts.device.getLocalId();
       excludedIds = [localFxADeviceId];
     }
     try {
@@ -564,13 +641,17 @@ ClientEngine.prototype = {
           prefName += "mobile";
           break;
         default:
-          this._log.warn(`Unexpected deviceType "${deviceType}" recording device telemetry.`);
+          this._log.warn(
+            `Unexpected deviceType "${deviceType}" recording device telemetry.`
+          );
           continue;
       }
       Services.telemetry.getHistogramById(hid).add(count);
       // Optimization: only write the pref if it changed since our last sync.
-      if (this._lastDeviceCounts == null ||
-          this._lastDeviceCounts.get(prefName) != count) {
+      if (
+        this._lastDeviceCounts == null ||
+        this._lastDeviceCounts.get(prefName) != count
+      ) {
         Svc.Prefs.set(prefName, count);
       }
     }
@@ -627,13 +708,18 @@ ClientEngine.prototype = {
   async handleHMACMismatch(item, mayRetry) {
     this._log.debug("Handling HMAC mismatch for " + item.id);
 
-    let base = await SyncEngine.prototype.handleHMACMismatch.call(this, item, mayRetry);
-    if (base != SyncEngine.kRecoveryStrategy.error)
+    let base = await SyncEngine.prototype.handleHMACMismatch.call(
+      this,
+      item,
+      mayRetry
+    );
+    if (base != SyncEngine.kRecoveryStrategy.error) {
       return base;
+    }
 
     // It's a bad client record. Save it to be deleted at the end of the sync.
     this._log.debug("Bad client record detected. Scheduling for deletion.");
-    this._deleteId(item.id);
+    await this._deleteId(item.id);
 
     // Neither try again nor error; we're going to delete it.
     return SyncEngine.kRecoveryStrategy.ignore;
@@ -646,14 +732,22 @@ ClientEngine.prototype = {
    * indicate higher importance.
    */
   _commands: {
-    resetAll:    { args: 0, importance: 0, desc: "Clear temporary local data for all engines" },
-    resetEngine: { args: 1, importance: 0, desc: "Clear temporary local data for engine" },
-    wipeAll:     { args: 0, importance: 0, desc: "Delete all client data for all engines" },
-    wipeEngine:  { args: 1, importance: 0, desc: "Delete all client data for engine" },
-    logout:      { args: 0, importance: 0, desc: "Log out client" },
-    displayURI:  { args: 3, importance: 1, desc: "Instruct a client to display a URI" },
-    repairRequest:  { args: 1, importance: 2, desc: "Instruct a client to initiate a repair" },
-    repairResponse: { args: 1, importance: 2, desc: "Instruct a client a repair request is complete" },
+    resetAll: {
+      args: 0,
+      importance: 0,
+      desc: "Clear temporary local data for all engines",
+    },
+    resetEngine: {
+      args: 1,
+      importance: 0,
+      desc: "Clear temporary local data for engine",
+    },
+    wipeEngine: {
+      args: 1,
+      importance: 0,
+      desc: "Delete all client data for engine",
+    },
+    logout: { args: 0, importance: 0, desc: "Log out client" },
   },
 
   /**
@@ -682,16 +776,26 @@ ClientEngine.prototype = {
       flowID: telemetryExtra.flowID,
     };
 
-    if ((await this._addClientCommand(clientId, action))) {
+    if (await this._addClientCommand(clientId, action)) {
       this._log.trace(`Client ${clientId} got a new action`, [command, args]);
-      this._tracker.addChangedID(clientId);
+      await this._tracker.addChangedID(clientId);
       try {
-        telemetryExtra.deviceID = this.service.identity.hashedDeviceID(clientId);
+        telemetryExtra.deviceID = this.service.identity.hashedDeviceID(
+          clientId
+        );
       } catch (_) {}
 
-      this.service.recordTelemetryEvent("sendcommand", command, undefined, telemetryExtra);
+      this.service.recordTelemetryEvent(
+        "sendcommand",
+        command,
+        undefined,
+        telemetryExtra
+      );
     } else {
-      this._log.trace(`Client ${clientId} got a duplicate action`, [command, args]);
+      this._log.trace(`Client ${clientId} got a duplicate action`, [
+        command,
+        args,
+      ]);
     }
   },
 
@@ -702,93 +806,47 @@ ClientEngine.prototype = {
    */
   async processIncomingCommands() {
     return this._notify("clients:process-commands", "", async function() {
-      if (!this.localCommands ||
-          (this._lastModifiedOnProcessCommands == this._localClientLastModified
-           && !this.ignoreLastModifiedOnProcessCommands)) {
+      if (
+        !this.localCommands ||
+        (this._lastModifiedOnProcessCommands == this._localClientLastModified &&
+          !this.ignoreLastModifiedOnProcessCommands)
+      ) {
         return true;
       }
       this._lastModifiedOnProcessCommands = this._localClientLastModified;
 
       const clearedCommands = await this._readCommands()[this.localID];
-      const commands = this.localCommands.filter(command => !hasDupeCommand(clearedCommands, command));
+      const commands = this.localCommands.filter(
+        command => !hasDupeCommand(clearedCommands, command)
+      );
       let didRemoveCommand = false;
-      let URIsToDisplay = [];
       // Process each command in order.
       for (let rawCommand of commands) {
         let shouldRemoveCommand = true; // most commands are auto-removed.
-        let {command, args, flowID} = rawCommand;
+        let { command, args, flowID } = rawCommand;
         this._log.debug("Processing command " + command, args);
 
-        this.service.recordTelemetryEvent("processcommand", command, undefined,
-                                          { flowID });
+        this.service.recordTelemetryEvent(
+          "processcommand",
+          command,
+          undefined,
+          { flowID }
+        );
 
         let engines = [args[0]];
         switch (command) {
           case "resetAll":
             engines = null;
-            // Fallthrough
+          // Fallthrough
           case "resetEngine":
             await this.service.resetClient(engines);
             break;
-          case "wipeAll":
-            engines = null;
-            // Fallthrough
           case "wipeEngine":
             await this.service.wipeClient(engines);
             break;
           case "logout":
             this.service.logout();
             return false;
-          case "displayURI":
-            let [uri, clientId, title] = args;
-            URIsToDisplay.push({ uri, clientId, title });
-            break;
-          case "repairResponse": {
-            // When we send a repair request to another device that understands
-            // it, that device will send a response indicating what it did.
-            let response = args[0];
-            let requestor = getRepairRequestor(response.collection);
-            if (!requestor) {
-              this._log.warn("repairResponse for unknown collection", response);
-              break;
-            }
-            if (!(await requestor.continueRepairs(response))) {
-              this._log.warn("repairResponse couldn't continue the repair", response);
-            }
-            break;
-          }
-          case "repairRequest": {
-            // Another device has sent us a request to make some repair.
-            let request = args[0];
-            let responder = getRepairResponder(request.collection);
-            if (!responder) {
-              this._log.warn("repairRequest for unknown collection", request);
-              break;
-            }
-            try {
-              if ((await responder.repair(request, rawCommand))) {
-                // We've started a repair - once that collection has synced it
-                // will write a "response" command and arrange for this repair
-                // request to be removed from the local command list - if we
-                // removed it now we might fail to write a response in cases of
-                // premature shutdown etc.
-                shouldRemoveCommand = false;
-              }
-            } catch (ex) {
-              if (Async.isShutdownException(ex)) {
-                // Let's assume this error was caused by the shutdown, so let
-                // it try again next time.
-                throw ex;
-              }
-              // otherwise there are no second chances - the command is removed
-              // and will not be tried again.
-              // (Note that this shouldn't be hit in the normal case - it's
-              // expected the responder will handle all reasonable failures and
-              // write a response indicating that it couldn't do what was asked.)
-              this._log.error("Failed to handle a repair request", ex);
-            }
-            break;
-          }
           default:
             this._log.warn("Received an unknown command: " + command);
             break;
@@ -800,11 +858,7 @@ ClientEngine.prototype = {
         }
       }
       if (didRemoveCommand) {
-        this._tracker.addChangedID(this.localID);
-      }
-
-      if (URIsToDisplay.length) {
-        this._handleDisplayURIs(URIsToDisplay);
+        await this._tracker.addChangedID(this.localID);
       }
 
       return true;
@@ -838,8 +892,14 @@ ClientEngine.prototype = {
       return;
     } else if (!args || args.length != commandData.args) {
       // Don't send a command with the wrong number of arguments.
-      this._log.error("Expected " + commandData.args + " args for '" +
-                      command + "', but got " + args);
+      this._log.error(
+        "Expected " +
+          commandData.args +
+          " args for '" +
+          command +
+          "', but got " +
+          args
+      );
       return;
     }
 
@@ -860,61 +920,9 @@ ClientEngine.prototype = {
     }
   },
 
-  /**
-   * Send a URI to another client for display.
-   *
-   * A side effect is the score is increased dramatically to incur an
-   * immediate sync.
-   *
-   * If an unknown client ID is specified, sendCommand() will throw an
-   * Error object.
-   *
-   * @param uri
-   *        URI (as a string) to send and display on the remote client
-   * @param clientId
-   *        ID of client to send the command to. If not defined, will be sent
-   *        to all remote clients.
-   * @param title
-   *        Title of the page being sent.
-   */
-  async sendURIToClientForDisplay(uri, clientId, title) {
-    this._log.trace("Sending URI to client: " + uri + " -> " +
-                   clientId + " (" + title + ")");
-    await this.sendCommand("displayURI", [uri, this.localID, title], clientId);
-
-    this._tracker.score += SCORE_INCREMENT_XLARGE;
-  },
-
-  /**
-   * Handle a bunch of received 'displayURI' commands.
-   *
-   * Interested parties should observe the "weave:engine:clients:display-uris"
-   * topic. The callback will receive an array as the subject parameter
-   * containing objects with the following keys:
-   *
-   *   uri       URI (string) that is requested for display.
-   *   clientId  ID of client that sent the command.
-   *   title     Title of page that loaded URI (likely) corresponds to.
-   *
-   * The 'data' parameter to the callback will not be defined.
-   *
-   * @param uris
-   *        An array containing URI objects to display
-   * @param uris[].uri
-   *        String URI that was received
-   * @param uris[].clientId
-   *        ID of client that sent URI
-   * @param uris[].title
-   *        String title of page that URI corresponds to. Older clients may not
-   *        send this.
-   */
-  _handleDisplayURIs: function _handleDisplayURIs(uris) {
-    Svc.Obs.notify("weave:engine:clients:display-uris", uris);
-  },
-
   async _removeRemoteClient(id) {
     delete this._store._remoteClients[id];
-    this._tracker.removeChangedID(id);
+    await this._tracker.removeChangedID(id);
     await this._removeClientCommands(id);
     this._modified.delete(id);
   },
@@ -944,14 +952,14 @@ ClientStore.prototype = {
   async createRecord(id, collection) {
     let record = new ClientsRec(collection, id);
 
-    const commandsChanges = this.engine._currentlySyncingCommands ?
-                            this.engine._currentlySyncingCommands[id] :
-                            [];
+    const commandsChanges = this.engine._currentlySyncingCommands
+      ? this.engine._currentlySyncingCommands[id]
+      : [];
 
     // Package the individual components into a record for the local client
     if (id == this.engine.localID) {
       try {
-        record.fxaDeviceId = await this.engine.fxAccounts.getDeviceId();
+        record.fxaDeviceId = await this.engine.fxAccounts.device.getLocalId();
       } catch (error) {
         this._log.warn("failed to get fxa device id", error);
       }
@@ -961,9 +969,15 @@ ClientStore.prototype = {
       record.protocols = SUPPORTED_PROTOCOL_VERSIONS;
 
       // Substract the commands we recorded that we've already executed
-      if (commandsChanges && commandsChanges.length &&
-          this.engine.localCommands && this.engine.localCommands.length) {
-        record.commands = this.engine.localCommands.filter(command => !hasDupeCommand(commandsChanges, command));
+      if (
+        commandsChanges &&
+        commandsChanges.length &&
+        this.engine.localCommands &&
+        this.engine.localCommands.length
+      ) {
+        record.commands = this.engine.localCommands.filter(
+          command => !hasDupeCommand(commandsChanges, command)
+        );
       }
 
       // Optional fields.
@@ -981,14 +995,18 @@ ClientStore.prototype = {
       // Add the commands we have to send
       if (commandsChanges && commandsChanges.length) {
         const recordCommands = record.cleartext.commands || [];
-        const newCommands = commandsChanges.filter(command => !hasDupeCommand(recordCommands, command));
+        const newCommands = commandsChanges.filter(
+          command => !hasDupeCommand(recordCommands, command)
+        );
         record.cleartext.commands = recordCommands.concat(newCommands);
       }
 
       if (record.cleartext.stale) {
         // It's almost certainly a logic error for us to upload a record we
         // consider stale, so make log noise, but still remove the flag.
-        this._log.error(`Preparing to upload record ${id} that we consider stale`);
+        this._log.error(
+          `Preparing to upload record ${id} that we consider stale`
+        );
         delete record.cleartext.stale;
       }
     }
@@ -1019,10 +1037,13 @@ ClientStore.prototype = {
       });
       let truncatedCommands = Utils.tryFitItems(commands, maxPayloadSize);
       if (truncatedCommands.length != record.commands.length) {
-        this._log.warn(`Removing commands from client ${id} (from ${record.commands.length} to ${truncatedCommands.length})`);
+        this._log.warn(
+          `Removing commands from client ${id} (from ${record.commands.length} to ${truncatedCommands.length})`
+        );
         // Restore original order.
-        record.commands = truncatedCommands.sort((a, b) =>
-          origOrder.get(a) - origOrder.get(b));
+        record.commands = truncatedCommands.sort(
+          (a, b) => origOrder.get(a) - origOrder.get(b)
+        );
       }
     }
     return record;
@@ -1035,8 +1056,9 @@ ClientStore.prototype = {
   async getAllIDs() {
     let ids = {};
     ids[this.engine.localID] = true;
-    for (let id in this._remoteClients)
+    for (let id in this._remoteClients) {
       ids[id] = true;
+    }
     return ids;
   },
 
@@ -1046,34 +1068,37 @@ ClientStore.prototype = {
 };
 
 function ClientsTracker(name, engine) {
-  Tracker.call(this, name, engine);
-  Svc.Obs.add("weave:engine:start-tracking", this);
-  Svc.Obs.add("weave:engine:stop-tracking", this);
+  LegacyTracker.call(this, name, engine);
 }
 ClientsTracker.prototype = {
-  __proto__: Tracker.prototype,
+  __proto__: LegacyTracker.prototype,
 
   _enabled: false,
 
-  observe: function observe(subject, topic, data) {
+  onStart() {
+    Svc.Obs.add("fxaccounts:new_device_id", this.asyncObserver);
+    Services.prefs.addObserver(
+      PREF_ACCOUNT_ROOT + "device.name",
+      this.asyncObserver
+    );
+  },
+  onStop() {
+    Services.prefs.removeObserver(
+      PREF_ACCOUNT_ROOT + "device.name",
+      this.asyncObserver
+    );
+    Svc.Obs.remove("fxaccounts:new_device_id", this.asyncObserver);
+  },
+
+  async observe(subject, topic, data) {
     switch (topic) {
-      case "weave:engine:start-tracking":
-        if (!this._enabled) {
-          Svc.Prefs.observe("client.name", this);
-          this._enabled = true;
-        }
-        break;
-      case "weave:engine:stop-tracking":
-        if (this._enabled) {
-          Svc.Prefs.ignore("client.name", this);
-          this._enabled = false;
-        }
-        break;
       case "nsPref:changed":
         this._log.debug("client.name preference changed");
-        this.addChangedID(this.engine.localID);
-        this.score += SCORE_INCREMENT_XLARGE;
+      // Fallthrough intended.
+      case "fxaccounts:new_device_id":
+        await this.addChangedID(this.engine.localID);
+        this.score += SINGLE_USER_THRESHOLD + 1; // ALWAYS SYNC NOW.
         break;
     }
-  }
+  },
 };
